
/*
 * General Purpose functions for the global management of the
 * Communication Processor Module.
 * Copyright (c) 1997 Dan Malek (dmalek@jlc.net)
 *
 * In addition to the individual control of the communication
 * channels, there are a few functions that globally affect the
 * communication processor.
 *
 * Buffer descriptors must be allocated from the dual ported memory
 * space.  The allocator for that is here.  When the communication
 * process is reset, we reclaim the memory available.  There is
 * currently no deallocator for this memory.
 * The amount of space available is platform dependent.  On the
 * MBX, the EPPC software loads additional microcode into the
 * communication processor, and uses some of the DP ram for this
 * purpose.  Current, the first 512 bytes and the last 256 bytes of
 * memory are used.  Right now I am conservative and only use the
 * memory that can never be used for microcode.  If there are
 * applications that require more DP ram, we can expand the boundaries
 * but then we have to be careful of any downloaded microcode.
 */
#include <linux/config.h>
#include <linux/errno.h>
#include <linux/sched.h>
#include <linux/kernel.h>
#include <linux/param.h>
#include <linux/string.h>
#include <linux/mm.h>
#include <linux/interrupt.h>
#include <linux/bootmem.h>
#include <asm/irq.h>
#include <asm/mpc8xx.h>
#include <asm/page.h>
#include <asm/pgtable.h>
#include <asm/8xx_immap.h>
#include "commproc.h"

typedef struct	cpm_dpalloc_entry_s {
	u_short				start_addr;
	u_short				size;
	struct	cpm_dpalloc_entry_s *	next_entr;
} cpm_dpalloc_entry_t;

static	void	cpm_dpalloc_init(void);
static	void	cpm_dpalloc_data_init(cpm_dpalloc_entry_t  *);
static	int	cpm_dpalloc_cl(cpm_dpalloc_entry_t *);
static	int	cpm_dpalloc_rem(cpm_dpalloc_entry_t **, cpm_dpalloc_entry_t *);
static	void	cpm_dpalloc_ins(cpm_dpalloc_entry_t **, cpm_dpalloc_entry_t *);

#define CPM_DPALLOC_COUNT	  (CPM_DATAONLY_SIZE / sizeof(cbd_t) + 1)  

/* Dynamic lists of allocated and free DPRAM 
*/
static	cpm_dpalloc_entry_t  *	cpm_dpalloc_free;
static	cpm_dpalloc_entry_t  *	cpm_dpalloc_alloc;
static	cpm_dpalloc_entry_t	cpm_dpalloc_list[CPM_DPALLOC_COUNT];

static	uint	host_buffer;	/* Start of the host buffer */
static	uint	host_end;	/* end + 1 */
cpm8xx_t	*cpmp;		/* Pointer to comm processor space */
				/* Number of pages in the host buffer */
 static	uint	host_page_num = (8
#ifdef CONFIG_SCC1_UART_RX_BDNUM
		+ CONFIG_SCC1_UART_RX_BDNUM * CONFIG_SCC1_UART_RX_BDSIZE
		+ CONFIG_SCC1_UART_TX_BDNUM * CONFIG_SCC1_UART_TX_BDSIZE
#endif
#ifdef CONFIG_SCC2_UART_RX_BDNUM
		+ CONFIG_SCC2_UART_RX_BDNUM * CONFIG_SCC2_UART_RX_BDSIZE
		+ CONFIG_SCC2_UART_TX_BDNUM * CONFIG_SCC2_UART_TX_BDSIZE
#endif
#ifdef CONFIG_SCC3_UART_RX_BDNUM
		+ CONFIG_SCC3_UART_RX_BDNUM * CONFIG_SCC3_UART_RX_BDSIZE
		+ CONFIG_SCC3_UART_TX_BDNUM * CONFIG_SCC3_UART_TX_BDSIZE
#endif
#ifdef CONFIG_SCC4_UART_RX_BDNUM
		+ CONFIG_SCC4_UART_RX_BDNUM * CONFIG_SCC4_UART_RX_BDSIZE
		+ CONFIG_SCC4_UART_TX_BDNUM * CONFIG_SCC4_UART_TX_BDSIZE
#endif
#ifdef CONFIG_SMC1_UART_RX_BDNUM
		+ CONFIG_SMC1_UART_RX_BDNUM * CONFIG_SMC1_UART_RX_BDSIZE
		+ CONFIG_SMC1_UART_TX_BDNUM * CONFIG_SMC1_UART_TX_BDSIZE
#endif
#ifdef CONFIG_SMC2_UART_RX_BDNUM
		+ CONFIG_SMC2_UART_RX_BDNUM * CONFIG_SMC2_UART_RX_BDSIZE
		+ CONFIG_SMC2_UART_TX_BDNUM * CONFIG_SMC2_UART_TX_BDSIZE
#endif
		- 1) / PAGE_SIZE + 1;

/* CPM interrupt vector functions.
*/
struct	cpm_action {
	void	(*handler)(void *);
	void	*dev_id;
};
static	struct	cpm_action cpm_vecs[CPMVEC_NR];
static	void	cpm_interrupt(int irq, void * dev, struct pt_regs * regs);
static	void	cpm_error_interrupt(void *);

void
m8xx_cpm_reset(void)
{
	volatile immap_t	*imp;
	volatile cpm8xx_t	*commproc;
	pte_t			*pte;
	int			i;

	imp = (immap_t *)IMAP_ADDR;
	commproc = (cpm8xx_t *)&imp->im_cpm;

#ifdef CONFIG_UCODE_PATCH
	/* Perform a reset.
	*/
	commproc->cp_cpcr = (CPM_CR_RST | CPM_CR_FLG);

	/* Wait for it.
	*/
	while (commproc->cp_cpcr & CPM_CR_FLG);

	cpm_load_patch(imp);
#endif

	/* Set SDMA Bus Request priority 5.
	 * On 860T, this also enables FEC priority 6.  I am not sure
	 * this is what we realy want for some applications, but the
	 * manual recommends it.
	 * Bit 25, FAM can also be set to use FEC aggressive mode (860T).
	*/
	imp->im_siu_conf.sc_sdcr = 1;

	/* Reclaim the DP memory for our use.
	*/
 	cpm_dpalloc_init();        

	/* Set the host page for allocation.
	*/
	host_buffer = (uint) alloc_bootmem_pages(PAGE_SIZE * host_page_num);
	host_end = host_buffer + PAGE_SIZE * host_page_num;
	for (i = 0; i < host_page_num; i++) {
		pte = va_to_pte(host_buffer + i * PAGE_SIZE);
		pte_val(*pte) |= _PAGE_NO_CACHE;
		flush_tlb_page(init_mm.mmap, host_buffer + i * PAGE_SIZE);
	}

	/* Tell everyone where the comm processor resides.
	*/
	cpmp = (cpm8xx_t *)commproc;
}

/* This is called during init_IRQ.  We used to do it above, but this
 * was too early since init_IRQ was not yet called.
 */
void
cpm_interrupt_init(void)
{
	/* Initialize the CPM interrupt controller.
	*/
	((immap_t *)IMAP_ADDR)->im_cpic.cpic_cicr =
	    (CICR_SCD_SCC4 | CICR_SCC_SCC3 | CICR_SCB_SCC2 | CICR_SCA_SCC1) |
		((CPM_INTERRUPT/2) << 13) | CICR_HP_MASK;
	((immap_t *)IMAP_ADDR)->im_cpic.cpic_cimr = 0;

	/* Set our interrupt handler with the core CPU.
	*/
	if (request_8xxirq(CPM_INTERRUPT, cpm_interrupt, 0, "cpm", NULL) != 0)
		panic("Could not allocate CPM IRQ!");

	/* Install our own error handler.
	*/
	cpm_install_handler(CPMVEC_ERROR, cpm_error_interrupt, NULL);
	((immap_t *)IMAP_ADDR)->im_cpic.cpic_cicr |= CICR_IEN;
}

/* CPM interrupt controller interrupt.
*/
static	void
cpm_interrupt(int irq, void * dev, struct pt_regs * regs)
{
	uint	vec;

	/* Get the vector by setting the ACK bit and then reading
	 * the register.
	 */
	((volatile immap_t *)IMAP_ADDR)->im_cpic.cpic_civr = 1;
	vec = ((volatile immap_t *)IMAP_ADDR)->im_cpic.cpic_civr;
	vec >>= 11;

	if (cpm_vecs[vec].handler != 0)
		(*cpm_vecs[vec].handler)(cpm_vecs[vec].dev_id);
	else
		((immap_t *)IMAP_ADDR)->im_cpic.cpic_cimr &= ~(1 << vec);

	/* After servicing the interrupt, we have to remove the status
	 * indicator.
	 */
	((immap_t *)IMAP_ADDR)->im_cpic.cpic_cisr = (1 << vec);
}

/* The CPM can generate the error interrupt when there is a race condition
 * between generating and masking interrupts.  All we have to do is ACK it
 * and return.  This is a no-op function so we don't need any special
 * tests in the interrupt handler.
 */
static	void
cpm_error_interrupt(void *dev)
{
}

/* Install a CPM interrupt handler.
*/
void
cpm_install_handler(int vec, void (*handler)(void *), void *dev_id)
{

	/* If null handler, assume we are trying to free the IRQ.
	*/
	if (!handler) {
		cpm_free_handler(vec);
		return;
	}

	if (cpm_vecs[vec].handler != 0)
		printk("CPM interrupt %x replacing %x\n",
			(uint)handler, (uint)cpm_vecs[vec].handler);
	cpm_vecs[vec].handler = handler;
	cpm_vecs[vec].dev_id = dev_id;
	((immap_t *)IMAP_ADDR)->im_cpic.cpic_cimr |= (1 << vec);
}

/* Free a CPM interrupt handler.
*/
void
cpm_free_handler(int vec)
{
	cpm_vecs[vec].handler = NULL;
	cpm_vecs[vec].dev_id = NULL;
	((immap_t *)IMAP_ADDR)->im_cpic.cpic_cimr &= ~(1 << vec);
}

/* Allocate some memory from the dual ported ram.  We may want to
 * enforce alignment restrictions, but right now everyone is a good
 * citizen.
 */
uint
m8xx_cpm_dpalloc(uint size)
{
	cpm_dpalloc_entry_t   * new_el;
	cpm_dpalloc_entry_t   * p;
	cpm_dpalloc_entry_t   * p1;
	uint			retloc;
	u_short			max;
	unsigned long		flags;
	int			i;

	size = (size + 7) & ~7;
	max = CPM_DATAONLY_SIZE;
	retloc = 0;
	new_el = NULL;

	if (size == 0) goto DONE;

	save_flags(flags);
	cli();

	/* Find free area in DPRAM
	*/
	for (p = cpm_dpalloc_free; p != NULL; p = p->next_entr) {
		if (p->size <= max && p->size >= size) {
			new_el = p;
			max = p->size;
		}
	}

	if (new_el == NULL) goto DONE1;

	/* Insert new element in the list of allocated DPRAM
	*/
	retloc = new_el->start_addr;
	p1 = cpm_dpalloc_list;
	p = NULL;
	i = 0;
	while (i < CPM_DPALLOC_COUNT) {
		if (p1->start_addr == 0 && !p1->size && !p1->next_entr) {
			p = p1;
			break;
		}
		i ++;
		p1 ++;
	}

	if (p == NULL) {
		panic ("m8xx_cpm_dpalloc: INTERNAL ERROR\n");
	}

	p->start_addr = new_el->start_addr;
	p->size = size;

	if (new_el->size > size) {
		new_el->size -= size;
		new_el->start_addr += size;
		cpm_dpalloc_ins(&cpm_dpalloc_alloc, p);
	} else {
		cpm_dpalloc_ins(&cpm_dpalloc_alloc, p);
		i = cpm_dpalloc_rem(&cpm_dpalloc_free, new_el);

		if ( i == -1) {
			panic ("m8xx_cpm_dpalloc: INTERNAL ERROR\n");
		}

		cpm_dpalloc_data_init(new_el);
	}

DONE1:
	restore_flags(flags);
DONE:
	return retloc;
}

int
m8xx_cpm_dpfree(uint start_addr)
{
	cpm_dpalloc_entry_t   * r;
	int			retloc;
	unsigned long		flags;

	retloc = -1;

	if ((start_addr < CPM_DATAONLY_BASE) ||
	    (start_addr > CPM_DATAONLY_SIZE + CPM_DATAONLY_BASE)) {
		goto DONE;
	}

	save_flags(flags);
	cli();

	for (r = cpm_dpalloc_alloc;
	     (r != NULL) && (r->start_addr != start_addr);
	     r = r->next_entr)
		/* EMPTY */ ;

	if (r) {
		retloc = cpm_dpalloc_rem(&cpm_dpalloc_alloc, r);

		if (retloc == -1) {
			panic("m8xx_cpm_dpfree: INTERNAL ERROR\n");
		}
		cpm_dpalloc_ins(&cpm_dpalloc_free, r);
		retloc = cpm_dpalloc_cl(cpm_dpalloc_free);

		if (retloc == -1) {
			panic("m8xx_cpm_dpfree: INTERNAL ERROR\n");
		}
		retloc = 0;
	}

	restore_flags(flags);
DONE:
	return retloc;
}

/* We also own one page of host buffer space for the allocation of
 * UART "fifos" and the like.
 */
uint
m8xx_cpm_hostalloc(uint size)
{
	uint	retloc;

	if ((host_buffer + size) >= host_end)
		return(0);

	retloc = host_buffer;
	host_buffer += size;

	return(retloc);
}

/* ** HG */
/*
 * Deallocate all host buffer ram
 */
void
m8xx_cpm_hostfree(void)
{
    /* Reset the host page for allocation.
    */
    host_buffer = host_end - PAGE_SIZE;
}
/* ** HG */

/* Set a baud rate generator.  This needs lots of work.  There are
 * four BRGs, any of which can be wired to any channel.
 * The internal baud rate clock is the system clock divided by 16.
 * This assumes the baudrate is 16x oversampled by the uart.
 */
#define BRG_INT_CLK		(((bd_t *)__res)->bi_intfreq * 1000000)
#define BRG_UART_CLK		(BRG_INT_CLK/16)
#define BRG_UART_CLK_DIV16	(BRG_UART_CLK/16)

void
m8xx_cpm_setbrg(uint brg, uint rate)
{
	volatile uint	*bp;

	/* This is good enough to get SMCs running.....
	*/
	bp = (uint *)&cpmp->cp_brgc1;
	bp += brg;
	/* The BRG has a 12-bit counter.  For really slow baud rates (or
	 * really fast processors), we may have to further divide by 16.
	 */
	if (((BRG_UART_CLK / rate) - 1) < 4096)
		*bp = (((BRG_UART_CLK / rate) - 1) << 1) | CPM_BRG_EN;
	else
		*bp = (((BRG_UART_CLK_DIV16 / rate) - 1) << 1) |
						CPM_BRG_EN | CPM_BRG_DIV16;
}

/* Initialize dynamic lists
 */
static void
cpm_dpalloc_init(void)
{
	int i;

	cpm_dpalloc_free = cpm_dpalloc_list;
	cpm_dpalloc_free->start_addr = CPM_DATAONLY_BASE;
	cpm_dpalloc_free->size = CPM_DATAONLY_SIZE;
	cpm_dpalloc_free->next_entr = NULL;
	cpm_dpalloc_alloc = NULL;

	for(i = 1; i <= CPM_DPALLOC_COUNT; i++) {
		cpm_dpalloc_data_init (cpm_dpalloc_list + i);
	}
}

/* Initialize element in list of DPRAM
 */
static void
cpm_dpalloc_data_init(cpm_dpalloc_entry_t * ptr)
{
	ptr->start_addr = 0;
	ptr->size = 0;
	ptr->next_entr = NULL;
}

static int
cpm_dpalloc_cl(cpm_dpalloc_entry_t * head)
{
	cpm_dpalloc_entry_t	* curr;
	cpm_dpalloc_entry_t	* next;
	int			  retloc;

	curr = head;
	retloc = -1;

	if (curr) {
		retloc = 0;

		while (next = curr->next_entr) {
			if (curr->start_addr + curr->size == next->start_addr) {
				curr->size = curr->size + next->size;
				curr->next_entr = next->next_entr;
				cpm_dpalloc_data_init(next);
			} else {
				curr = next;
			}
		 }
	}

	return retloc;
}

/* Remove element from dynamic list of DPRAM
 */
static int
cpm_dpalloc_rem(cpm_dpalloc_entry_t ** head, cpm_dpalloc_entry_t * ptr)
{
	cpm_dpalloc_entry_t	* prev;
	cpm_dpalloc_entry_t	* curr;
	int			  retloc;

	retloc = -1;
	for (prev = NULL , curr = *head;
	     curr != NULL && ptr != curr;
	     prev = curr , curr = curr->next_entr)
		/* EMPTY */ ;

	if (curr) {
		if (prev) {
			prev->next_entr = curr->next_entr;
		} else {
			*head = curr->next_entr;
		}
		retloc = 0;
	}

	return retloc;
}

/* Insert element in dynamic list of DPRAM
 */
static void
cpm_dpalloc_ins(cpm_dpalloc_entry_t ** head, cpm_dpalloc_entry_t * ptr)
{
	cpm_dpalloc_entry_t	* prev;
	cpm_dpalloc_entry_t	* curr;

	for (prev = NULL , curr = *head;
	     curr != NULL && ptr->start_addr >= curr->start_addr;
	     prev = curr , curr = curr->next_entr)
		/* EMPTY */ ;

	ptr->next_entr = curr;

	if (prev) {
		prev->next_entr = ptr;
	} else {
		*head = ptr;
	}
}


#ifdef CONFIG_8XX_USE_RISC_TIMERS
/*
 * Disable RISC Timer Tick, and issue dummy command so that we can be
 * sure that the CPM has really completed any possibly running RISC
 * timer updates.
 *
 * NOTE: THIS CODE ASSUMES THAT RISC TIMER #15 IS never USED!
 */
static int m8xx_pbdat_synchronize_cpm (volatile immap_t *immr)
{
	volatile rt_pram_t *rtp = (rt_pram_t *)&cpmp->cp_dparam[PROFF_RTMR];
	unsigned long flags;

	/* Check if RISC timers are ticking... */
	if ((immr->im_cpm.cp_rccr & RCCR_TIME) == 0)
		return (0);
	
	save_flags(flags);
	cli();

	/* disable RISC timer tick */
	immr->im_cpm.cp_rccr &= ~(RCCR_TIME);

	/* setup disable command for RISC timer # 15 */
	rtp->tm_cmd = TM_CMD_NUM(15);	/* TM_CMD_VALID is NOT set */

	/* Issue SET TIMER command to CPM */
	immr->im_cpm.cp_cpcr = mk_cr_cmd (CPM_CR_CH_TIMER, CPM_CR_SET_TIMER) | CPM_CR_FLG;
	while (immr->im_cpm.cp_cpcr & CPM_CR_FLG)
		;

	restore_flags(flags);

	return (1);
}

void m8xx_pbdat_set_mask  (volatile immap_t *immr, unsigned long mask)
{
	int flag = m8xx_pbdat_synchronize_cpm (immr);

	immr->im_cpm.cp_pbdat |= mask;

	/* re-enable RISC timers if they were on */
	if (flag) {
		immr->im_cpm.cp_rccr |= RCCR_TIME;
	}
}

void m8xx_pbdat_clear_mask(volatile immap_t *immr, unsigned long mask)
{
	int flag = m8xx_pbdat_synchronize_cpm (immr);

	immr->im_cpm.cp_pbdat &= ~(mask);

	/* re-enable RISC timers if they were on */
	if (flag) {
		immr->im_cpm.cp_rccr |= RCCR_TIME;
	}
}

#endif /* CONFIG_8XX_USE_RISC_TIMERS */
