/***********************************************************************
 *
 * (C) Copyright 2001, 2002
 * DENX Software Engineering
 * Wolfgang Denk, wd@denx.de
 * All rights reserved.
 *
 ***********************************************************************/

#include <linux/kernel.h>
#include <linux/init.h>
#include <linux/module.h>
#include <linux/spinlock.h>
#include <linux/smp_lock.h>	/* For (un)lock_kernel */
#include <linux/input.h>	/* interface to input subsystem */
#include <linux/irq.h>
#include <linux/poll.h>
#include <asm/uaccess.h>
#include <asm/io.h>
#include <asm/semaphore.h>
#include <asm/8xx_immap.h>
#include <asm/mpc8xx.h>
/* #include <asm/commproc.h> -- not yet */
#include "commproc.h"

#include <linux/interrupt.h>

#include <../drivers/i2c/i2c-simple.h>


#undef	DEBUG

#ifdef	DEBUG
# define debugk(fmt,args...)	printk(fmt ,##args)
#else
# define debugk(fmt,args...)
#endif

/*
 * For character devices
 */
#include <linux/fs.h>		/* character device definitions */
#include <linux/wrapper.h>	/* wrapper for compatibility with future versions */


#ifndef KBD_MAJOR
#define KBD_MAJOR	11	/* Raw Keyboard Device	*/
#endif	/* KBD_MAJOR */


/*
 * We implement 2 minor devices:
 *
 * min 0 provides "normal" keyboard data which can be read by any
 *       tool like "cat" or "xd" etc. For each key press exactly one
 *       character can be read from the device. Because of the
 *       special requirements of the keybrd control application no
 *       autorepeat is implemented here.
 *
 * min 1 provides a special interface for the contol application. It
 *       is optimized to provide exact timing information for the
 *       duration of each key press. For each "key event" _two_ bytes
 *       of data can be read from the device: the keycode, and a byte
 *       of "modifiers" which includes information about SHIFT and
 *       CONTROL status and if this was a key PRESS or a key RELEASE
 *       event. For each key press two events are generated: the
 *       first one when the key is pressed, and the second one when
 *       it is released.
 *
 * Warning: both minor devices read from the same input buffer; do
 * not try to read from both devices simultanously.
 */
#define KBD_MINOR	2	/* allow for 2 minor devices */
int	kbd_dev_flags[KBD_MINOR];

#define KBD_DEV_OPEN	1

static int	kbd_open     (struct inode *, struct file *);
static int	kbd_release  (struct inode *, struct file *);
static ssize_t	kbd_read     (struct file *, char *, size_t, loff_t *);
static unsigned int kbd_poll (struct file *, poll_table *);

static struct file_operations kbd_fops = {
	owner:		THIS_MODULE,
	open:		kbd_open,
	release:	kbd_release,
	read:		kbd_read,
	poll:		kbd_poll,
};

/*
 * Keyboard Controller
 *
 * command codes
 */
#define KEYBD_CMD_READ_KEYS	0x01
#define KEYBD_CMD_READ_VERSION	0x02
#define KEYBD_CMD_READ_STATUS	0x03
#define KEYBD_CMD_RESET_ERRORS	0x10

/* Number of bytes returned from Keyboard Controller */
#define KEYBD_VERSIONLEN	2	/* version information */
#define KEYBD_DATALEN		9	/* normal key scan data */

/*
 * Input ring buffer
 * Holds two bytes of data per key event (modifiers + keycode).
 * first == last ==> RB empty.
 * first == (last+1) mod size ==> RB full.
 */
#define RB_SIZE 256			/* Ring buffer size */
static unsigned char	rb_data[2*RB_SIZE];
static volatile unsigned short	rb_first, rb_last;
static spinlock_t rb_lock;
static DECLARE_WAIT_QUEUE_HEAD(rb_wait);
static void rb_add (unsigned char keycd, unsigned char mod);
static int  rb_get (unsigned int n, unsigned char *buf, int flag);


typedef struct {
	struct task_struct *thread;	/* Linux task structure of thread */
	struct semaphore    sem;	/* semaphore for thread control */
	wait_queue_head_t   queue;	/* queue thread is waiting on */
	volatile unsigned int pend;	/* key event pending */
} kbd_dev_t;

kbd_dev_t kbd_dev;

/*
 * The name for our device, as it will appear in /proc/devices
 */
#define DEVICE_NAME "lwmon_kbd"

/*
 * We implement 2 minor devices:
 *
 * minor 0: This device will be used by	 the  LICCON  software	which
 *	    needs  exact  timing  information  since keys are used to
 *	    control hardware;  each  key  event	 (key  press  or  key
 *	    release)  will  be reported to the application as 2 bytes
 *	    of data: the first byte contains key status	 data  (press
 *	    or	release	 event, shift, control and error status), and
 *	    the second byte the decoded key code.
 *
 * minor 1: This will be used mainly for testing and in case  a	 more
 *	    conventional  application  needs  to access the keyboard:
 *	    only the decoded keycodes are sent to the application.
 *
 * The keyboard driver buffers key events if  needed,  but  then  all
 * timing information gets lost; also, no key repeat is performed.
 *
 * Implemented Keyboard Map:
 *
 *     Key   Unshifted	Shift	Ctrl  Shift+Ctrl      HW-Code	Algo.
 *     --------------------------------------------------------------
 *	F1	0x80	0xA0	0x90	0xB0		0x3A	0
 *	F2	0x81	0xA1	0x91	0xB1		0x3B	0
 *	F3	0x82	0xA2	0x92	0xB2		0x3C	0
 *	F4	0x83	0xA3	0x93	0xB3		0x3D	0
 *	F5	0x84	0xA4	0x94	0xB4		0x3E	0
 *	F6	0x85	0xA5	0x95	0xB5		0x3F	0
 *	F7	0x86	0xA6	0x96	0xB6		0x40	0
 *	F8	0x87	0xA7	0x97	0xB7		0x41	0
 *
 *	<	0x71	0x51	0x11	0xF1		0x50	1
 *	>	0x72	0x52	0x12	0xF2		0x4F	1
 *
 *	ENTER	0x0D	0xAD	0x9D	0xBD		0x28	4
 *
 *	P0	0x6D	0x4D	0x0D	0xED		0x10	1
 *	P1	0x61	0x41	0x01	0xE1		0x04	1
 *	P2	0x62	0x42	0x02	0xE2		0x05	1
 *	P3	0x6C	0x4C	0x0C	0xEC		0x0F	1
 *	P4	0x64	0x44	0x04	0xE4		0x08	1
 *	P5	0x65	0x45	0x05	0xE5		0x09	1
 *	P6	0x66	0x46	0x06	0xE6		0x07	1
 *	P7	0x73	0x53	0x13	0xF3		0x16	1
 *	P8	0x74	0x54	0x14	0xF4		0x17	1
 *
 *	P	0x70	0x50	0x10	0xF0		0x13	1
 *	.	0x2E	0x3E	0x1E	0xFE		0x37	2
 *
 *	0	0x30	0x20	0x00	0xD0		0x27	3
 *	1	0x31	0x21	0x01	0xD1		0x1E	3
 *	2	0x32	0x22	0x02	0xD2		0x1F	3
 *	3	0x33	0x23	0x03	0xD3		0x20	3
 *	4	0x34	0x24	0x04	0xD4		0x21	3
 *	5	0x35	0x25	0x05	0xD5		0x22	3
 *	6	0x36	0x26	0x06	0xD6		0x23	3
 *	7	0x37	0x27	0x07	0xD7		0x24	3
 *	8	0x38	0x28	0x08	0xD8		0x25	3
 *	9	0x39	0x29	0x09	0xD9		0x26	3
 *
 * Unfortunately, this keyboard map is VERY  irregular;	 to  minimize
 * memory footprint (for code tables) and CPU (for decoding) we store
 * a  key map that maps the HW code (as read from the PIC controller)
 * into the _Shift+Ctrl_ key  code,  plus  a  "algorithm  code".  The
 * following algorithms are used:
 *
 * Algo 0:	Unshifted = Shift+Ctrl - 0x30
 *		Shift	  = Shift+Ctrl - 0x10
 *		Ctrl	  = Shift+Ctrl - 0x20
 *
 *		Used for: F1 F2 F3 F4 F5 F6 F7 F8
 *
 * Algo 1:	Unshifted = Shift+Ctrl - 0x80
 *		Shift	  = Shift+Ctrl - 0xA0
 *		Ctrl	  = Shift+Ctrl - 0xE0
 *
 *		Used for: < > P0 P1 P2 P3 P4 P5 P6 P7 P8 P
 *
 * Algo 2:	Unshifted = Shift+Ctrl - 0xD0
 *		Shift	  = Shift+Ctrl - 0xC0
 *		Ctrl	  = Shift+Ctrl - 0xE0
 *
 *		Used for: .
 *
 * Algo 3:	Unshifted = Shift+Ctrl - 0xA0
 *		Shift	  = Shift+Ctrl - 0xB0
 *		Ctrl	  = Shift+Ctrl - 0xD0
 *
 *		Used for: 0 1 2 3 4 5 6 7 8 9
 *
 * Algo 4:	Unshifted = Shift+Ctrl - 0xB0
 *		Shift	  = Shift+Ctrl - 0x10
 *		Ctrl	  = Shift+Ctrl - 0x20
 *
 *		Used for: ENTER
 *
 */

#define LWMON_KBD_MAX_MINOR	2


static unsigned char kbd_map [] = {
/*	   HW Code  Shift+Ctrl Algo        Key	*/
	[ 2 * 0x04 ] = 0xE1,	1,	/* P1	*/
	[ 2 * 0x05 ] = 0xE2,	1,	/* P2	*/
	[ 2 * 0x07 ] = 0xE6,	1,	/* P6	*/
	[ 2 * 0x08 ] = 0xE4,	1,	/* P4	*/
	[ 2 * 0x09 ] = 0xE5,	1,	/* P5	*/
	[ 2 * 0x0F ] = 0xEC,	1,	/* P3	*/
	[ 2 * 0x10 ] = 0xED,	1,	/* P0	*/
	[ 2 * 0x13 ] = 0xF0,	1,	/* P	*/
	[ 2 * 0x16 ] = 0xF3,	1,	/* P7	*/
	[ 2 * 0x17 ] = 0xF4,	1,	/* P8	*/
	[ 2 * 0x1E ] = 0xD1,	3,	/* 1	*/
	[ 2 * 0x1F ] = 0xD2,	3,	/* 2	*/
	[ 2 * 0x20 ] = 0xD3,	3,	/* 3	*/
	[ 2 * 0x21 ] = 0xD4,	3,	/* 4	*/
	[ 2 * 0x22 ] = 0xD5,	3,	/* 5	*/
	[ 2 * 0x23 ] = 0xD6,	3,	/* 6	*/
	[ 2 * 0x24 ] = 0xD7,	3,	/* 7	*/
	[ 2 * 0x25 ] = 0xD8,	3,	/* 8	*/
	[ 2 * 0x26 ] = 0xD9,	3,	/* 9	*/
	[ 2 * 0x27 ] = 0xD0,	3,	/* 0	*/
	[ 2 * 0x28 ] = 0xBD,	4,	/* ENTER*/
	[ 2 * 0x37 ] = 0xFE,	2,	/* .	*/
	[ 2 * 0x3A ] = 0xB0,	0,	/* F1	*/
	[ 2 * 0x3B ] = 0xB1,	0,	/* F2	*/
	[ 2 * 0x3C ] = 0xB2,	0,	/* F3	*/
	[ 2 * 0x3D ] = 0xB3,	0,	/* F4	*/
	[ 2 * 0x3E ] = 0xB4,	0,	/* F5	*/
	[ 2 * 0x3F ] = 0xB5,	0,	/* F6	*/
	[ 2 * 0x40 ] = 0xB6,	0,	/* F7	*/
	[ 2 * 0x41 ] = 0xB7,	0,	/* F8	*/
	[ 2 * 0x4F ] = 0xF2,	1,	/* >	*/
	[ 2 * 0x50 ] = 0xF1,	1,	/* <	*/
};

#define KBD_MOD_CONTROL		0x01		/* Control */
#define KBD_MOD_SHIFT		0x02		/* Shift */
#define KBD_MOD_PRESS		0x10		/* Key press(1)/release(0) */
#define KBD_MOD_ERROR		0x80		/* Error condition */

#define KBD_ERR_ROLLOVR		0x01		/* Error RollOver */
#define KBD_ERR_POST		0x02		/* POST Fail */
#define KBD_ERR_UNDEF		0x03		/* Undefined Error */
#define KBD_ERR_MAX	(KBD_ERR_UNDEF + 1)

static unsigned char kbd_mod_state;
static unsigned char kbd_key_state [sizeof(kbd_map)/2];
static unsigned int nkey = sizeof(kbd_map) / 2;

/*
 * Prototypes
 */
static	int	lwmon_kbd_init (void) __init;

static	void	lwmon_readkeys	(void);
static unsigned char kbd_decode (unsigned char, unsigned char);
static	int	keyboard_thread (void *dev_arg);
static	void	setup_thread	(kbd_dev_t *dev);
static	void	leave_thread	(kbd_dev_t *dev);

int	init_module(void);
void	cleanup_module(void);

MODULE_AUTHOR("Wolfgang Denk, wd@denx.de");
MODULE_DESCRIPTION("Keyboard Driver for LWMON Project");

/* ==================================================================== */
static int kbd_open (struct inode *inode, struct file *file)
{
	int minor = MINOR(inode->i_rdev);

	debugk ("kbd_open: minor %d\n", minor);

	if (minor >= KBD_MINOR)
		return (-ENXIO);

	if (kbd_dev_flags[minor] & KBD_DEV_OPEN)
		return (-EBUSY);

	kbd_dev_flags[minor] |= KBD_DEV_OPEN;

	MOD_INC_USE_COUNT;

	return (0);
}

static int kbd_release (struct inode *inode, struct file *file)
{
	int minor = MINOR(inode->i_rdev);

	debugk ("kbd_release: minor %d\n", minor);

	if (minor >= KBD_MINOR)
		return (-ENXIO);

	kbd_dev_flags[minor] &= ~KBD_DEV_OPEN;

	MOD_DEC_USE_COUNT;

	return (0);
}

static ssize_t kbd_read (struct file *file,
			 char *buf, size_t count, loff_t *ppos)
{
	DECLARE_WAITQUEUE(wait, current);
	unsigned char io_buf[2*RB_SIZE];
	int rc = 0;
	int len = 0;
	int minor;
	unsigned int n;

	minor = MINOR(file->f_dentry->d_inode->i_rdev);


	if (minor == 1) {
		/* Two bytes per key event */
		if ((count % 2) != 0) /* must always read a multiple of 2 */
			return (-EINVAL);
		count /= 2;
	}

	n = (count > RB_SIZE) ? RB_SIZE : count;

	add_wait_queue(&rb_wait, &wait);

	for (;;) {

		set_current_state (TASK_INTERRUPTIBLE);

		if ((len = rb_get (n, io_buf, minor)) != 0) {
			break;
		}

		if (file->f_flags & O_NONBLOCK) {
			rc = -EAGAIN;
			goto OUT;
		}

		if (signal_pending(current)) {
			rc = -ERESTARTSYS;
			goto OUT;
		}

		schedule ();
	}

	/* Copy out */
	if ((rc = verify_area(VERIFY_WRITE, buf, len)) != 0) {
		goto OUT;
	}

	copy_to_user((void *)buf, (void*)(&io_buf[0]), len);
	rc = len;
OUT:
	set_current_state (TASK_RUNNING);
	remove_wait_queue(&rb_wait, &wait);

	debugk ("kbd_read: rc=%d\n", rc);

	return rc;
}

static unsigned int kbd_poll (struct file *file, poll_table *wait)
{
	debugk ("kbd_poll enter\n");

	poll_wait(file, &rb_wait, wait);

	debugk ("kbd_poll rc=%d\n", (rb_first != rb_last) ? 1 : 0);

	return (rb_first != rb_last) ? POLLIN | POLLRDNORM : 0;
}

/* ==================================================================== */
/* ==================================================================== */

/*
 * LWMON Keyboard controller interrupt
 */
static void lwmon_kbd_interrupt (int irq, void *dev_arg, struct pt_regs * regs)
{
	kbd_dev_t *dev = (kbd_dev_t *)dev_arg;

	debugk ("KBD INT\n");
	dev->pend = 1;
	wake_up_interruptible(&dev->queue);
}


/*
 * Initialize the driver - Register the character device
 */
static int __init lwmon_kbd_init (void)
{
	int rc;
	si2c_request_t i2c_rq;

	debugk (__FUNCTION__ "() called\n");

	printk (KERN_INFO "Kbd Driver: $Revision: 1.1.1.1 $ wd@denx.de\n");

	/*
	 * Verify correct operation of keyboard controller
	 */
	i2c_rq.alen = 1;
	i2c_rq.addr[0] = I2C_ADDR_KEYBD;

	i2c_rq.dlen = 1;
	i2c_rq.data[0] = KEYBD_CMD_READ_STATUS;

	(void) si2c_write (&i2c_rq);
	(void) si2c_read (&i2c_rq);
	debugk ("Keybd status code: 0x%02X\n", i2c_rq.data[0]);
	if (i2c_rq.data[0]) {
		printk (" - aborted: error code %02X\n", i2c_rq.data[0]);
		return (-EIO);
	}

	/* read kbd controller version */
	i2c_rq.dlen = 1;
	i2c_rq.data[0] = KEYBD_CMD_READ_VERSION;
	(void) si2c_write (&i2c_rq);
	i2c_rq.dlen = KEYBD_VERSIONLEN;
	(void) si2c_read (&i2c_rq);
	printk (KERN_INFO "Kbd Controller Version %d.%d\n",
		i2c_rq.data[0], i2c_rq.data[1]);

	init_MUTEX_LOCKED (&kbd_dev.sem);

	kernel_thread (keyboard_thread, (void*)&kbd_dev, 0);

	/* wait till it has reached the setup_thread routine */
	down (&kbd_dev.sem);

	/* Install Interrupt handler, edge triggered */
	((immap_t *) IMAP_ADDR)->im_siu_conf.sc_siel |=
		(0x80000000 >> KBD_INTERRUPT);

	rc = request_8xxirq (KBD_INTERRUPT, lwmon_kbd_interrupt, 0, "kbd",
				(void *)&kbd_dev);

	if (rc != 0) {
		printk ("Can't install KBD IRQ %d (rc=%d)\n",
			KBD_INTERRUPT, rc);
		return (rc);
	}

	debugk ("Installed KBD IRQ %d\n", KBD_INTERRUPT);

	rc = register_chrdev(KBD_MAJOR, DEVICE_NAME, &kbd_fops);

	if (rc != 0) {
		free_irq (KBD_INTERRUPT, (void *)&kbd_dev);
		return (rc);
	}
	return (0);
}

/******************************
 **** Module Declarations *****
 **************************** */

module_init (lwmon_kbd_init);

#ifdef MODULE
/*
 * Cleanup - unregister the driver
 */
static void lwmon_kbd_cleanup (void)
{
	int rc;

	debugk (__FUNCTION__ "() called\n");

	if (kbd_dev.thread == 0) {
		printk ("Can't terminate non-existant LWMON keyboard thread\n");
	} else {
		init_MUTEX_LOCKED (&kbd_dev.sem);

		kill_proc (kbd_dev.thread->pid, SIGKILL, 1);

		/* block till thread terminated */
		down (&kbd_dev.sem);
	}

	free_irq (KBD_INTERRUPT, (void *)&kbd_dev);
	debugk ("Freed KBD IRQ %d\n", KBD_INTERRUPT);

	rc = unregister_chrdev(KBD_MAJOR, DEVICE_NAME);

	if (rc < 0) {
		printk ("lwmon_kbd_cleanup: unregister_chrdev rc=%d\n", rc);
	}
}

module_exit (lwmon_kbd_cleanup);

#endif	/* MODULE */


static void lwmon_readkeys (void)
{
	ssize_t n;
	int i, j;
	si2c_request_t i2c_rq;
	unsigned char c, mod;

	debugk (__FUNCTION__ "() called\n");

	i2c_rq.alen = 1;
	i2c_rq.addr[0] = I2C_ADDR_KEYBD;

	/* read key codes */
	i2c_rq.dlen = 1;
	i2c_rq.data[0] = KEYBD_CMD_READ_KEYS;
	(void) si2c_write (&i2c_rq);
	i2c_rq.dlen = KEYBD_DATALEN;
	n = si2c_read (&i2c_rq);

#ifdef DEBUG
	printk ("KEYS:");
	for (i=0; i<n; ++i) {
		printk (" %02X", i2c_rq.data[i]);
	}
	printk ("\n");
#endif

	/*
	 * Handle error conditions: bytes 1 ... N will contain error
	 * codes (0 < error code < KBD_ERR_MAX), so it should be
	 * sufficient to test the first byte.
	 */
	if ((c = i2c_rq.data[1]) != 0) {
		if (c < KBD_ERR_MAX) {
			printk (KERN_INFO "KBD: error code %d\n", c);
			return;
		}
		if (c >= nkey) {
			printk (KERN_INFO "KBD: invalid key code 0x%02x\n", c);
			return;
		}
	}

	/*
	 * Handle modifiers
	 * These are encoded by bits in the first data byte
	 */
	mod = i2c_rq.data[0];

	if ((c = mod & KBD_MOD_CONTROL) ^ (kbd_mod_state & KBD_MOD_CONTROL)) {
		kbd_mod_state = (kbd_mod_state & ~KBD_MOD_CONTROL) | c;
		debugk ("KEY: CNTL %s\n", (!!c) ? "pressed" : "released");
	}
	if ((c = mod & KBD_MOD_SHIFT) ^ (kbd_mod_state & KBD_MOD_SHIFT)) {
		kbd_mod_state = (kbd_mod_state & ~KBD_MOD_SHIFT) | c;
		debugk ("KEY: SHIFT %s\n", (!!c) ? "pressed" : "released");
	}

	/*
	 * Scan for released and pressed keys
	 *
	 * Optimization: the i2c_rq.data[] array is filled in "from
	 * left", i. e. we can stop searching after the first empty
	 * key entry
	 *
	 * Optimization: the i2c_rq.data[] array will not contain
	 * duplicate entries for the same key code thus it's OK
	 * to perform a single linear scan though all key codes
	 */
	for (i=0; i<nkey; ++i) {
		unsigned char found = 0;

		if (kbd_map[i+i] == 0) {	/* impossible key code */
			continue;
		}

		c = kbd_key_state[i];		/* previous state of this key */

		for (j=1; j<n && i2c_rq.data[j]; j++) {
			if (i == i2c_rq.data[j]) {	/* key "i" pressed */
				unsigned char keycd;

				found = 1;
				if (c) {
					continue;	/* no change */
				}
				kbd_key_state[i] = 1;

				keycd = kbd_decode (i, kbd_mod_state);

				debugk ("KEY: %02x (%c) pressed\n",
					keycd,
					(keycd>=' ' && keycd<0x80) ? keycd : ' ');

				rb_add (keycd, kbd_mod_state | KBD_MOD_PRESS);
				break;		/* this key won't be there twice */
			}
		}
		if (c && !found) {		/* key "i" released */
			unsigned char keycd;
			kbd_key_state[i] = 0;

			keycd = kbd_decode (i, kbd_mod_state);

			debugk ("KEY: %02x (%c) released\n",
				keycd,
				(keycd>=' ' && keycd<0x80) ? keycd : ' ');

			rb_add (keycd, kbd_mod_state);
		}
	}
}

static
unsigned char kbd_decode (unsigned char hw_code, unsigned char mod)
{
	unsigned char code = kbd_map [ 2 * hw_code ];
	unsigned char algo = kbd_map [ 2 * hw_code +1 ];

	if (hw_code >= nkey) {
		debugk ("decode: illegal hw_code %02X [max. %02X]\n",
			hw_code, nkey);
		return 0;
	}

	code = kbd_map [ 2 * hw_code ];
	algo = kbd_map [ 2 * hw_code +1 ];

	if (code == 0) {
		debugk ("decode: illegal keycode\n");
		return 0;	/* illegal keycode - can't happen */
	}

#if 0
	debugk ("decode: %02x:%02x => algo %d code %02X\n",
		mod, hw_code, algo, code);
#endif

	switch (mod & (KBD_MOD_SHIFT | KBD_MOD_CONTROL)) {
	case 0:				/* neither shift nor control	*/
		switch (algo) {
		case 0:	return (code - 0x30);
		case 1: return (code - 0x80);
		case 2: return (code - 0xD0);
		case 3: return (code - 0xA0);
		case 4: return (code - 0xB0);
		}
		return 0;
	case KBD_MOD_SHIFT:				/* shift only	*/
		switch (algo) {
		case 0:	/* fall through */
		case 4: return (code - 0x10);
		case 1: return (code - 0xA0);
		case 2: return (code - 0xC0);
		case 3: return (code - 0xB0);
		}
		return 0;
	case KBD_MOD_CONTROL:				/* control only	*/
		switch (algo) {
		case 0:	/* fall through */
		case 4: return (code - 0x20);
		case 1: /* fall through */
		case 2: return (code - 0xE0);
		case 3: return (code - 0xD0);
		}
		return 0;
	case (KBD_MOD_SHIFT | KBD_MOD_CONTROL):		/* both		*/
		return (code);
	}
	return 0;
}

static void rb_add (unsigned char keycd, unsigned char mod)
{
	unsigned short rb_curr, rb_next;

	spin_lock (&rb_lock);			/* grab lock */

	rb_next = (rb_last + 2) % (2*RB_SIZE);

	if (rb_next == rb_first) {
printk ("## RINGBUFFER FULL\n");
		goto DONE;
	}

	rb_curr = (rb_first == rb_last) ? rb_last : rb_next;

	rb_data[rb_curr  ] = keycd;
	rb_data[rb_curr+1] = mod;

	rb_last = rb_next;

	wake_up_interruptible(&rb_wait);
DONE:
	spin_unlock (&rb_lock);			/* release lock */

	debugk("RB_ADD: %d / %d [%02X / %02X]\n",
		rb_first, rb_last, keycd, mod);
}

static int rb_get (unsigned int n, unsigned char *buf, int flag)
{
	int cnt = 0;

	spin_lock (&rb_lock);			/* grab lock */

	while ((cnt <= n) && (rb_first != rb_last)) {
		unsigned char keycd, mod;

		keycd = rb_data[rb_first++];	/* copy keycode */
		mod   = rb_data[rb_first++];	/* copy mode */

		if (rb_first >= (2*RB_SIZE))
			rb_first = 0;

		if (flag == 0) {	/* copy only key PRESS events */
			if (mod & KBD_MOD_PRESS) {
				*buf++ = keycd;
				cnt++;
			}
		} else {		/* copy everything */
			*buf++ = keycd;
			*buf++ = mod;
			cnt += 2;
		}
	}

	spin_unlock (&rb_lock);			/* release lock */

	return (cnt);
}
/******************************
 ****	Keyboard Thread	  *****
 **************************** */

static int keyboard_thread (void *dev_arg)
{
	kbd_dev_t *dev = (kbd_dev_t *)dev_arg;

	setup_thread (dev);		/* setup the thread environment */

	debugk ("KBD Thread started\n");

	lwmon_readkeys ();		/* read any stuck keys, and enable interrupts */

	for (;;) {
		/*
		 * wait for interrupt
		 */
		if (wait_event_interruptible(dev->queue, dev->pend)) {
			break;
		}

		/*
		 * We cannot allow new interrupts from the keyboard
		 * controller before we have not completed to read
		 * (and thus ACK) the keypress over the I2C bus...
		 */
		mask_irq(KBD_INTERRUPT);
		lwmon_readkeys ();
		dev->pend = 0;
		unmask_irq(KBD_INTERRUPT);
	}

	leave_thread (dev);

	return (0);
}

static void setup_thread (kbd_dev_t *dev)
{
	/* lock the kernel */
	lock_kernel();

	/* release insmod's memory map, use the kernel one's */
	exit_mm(current);

	/* set session leader and process group to init process */
	current->session = 1;
	current->pgrp = 1;

	/* set name of this process (max 15 chars + 0 !) */
	strcpy (current->comm, "LWMON keyboard");

	/* disconnect from the fs usage */
	exit_files (current);
	exit_fs	   (current);

	/* fill in thread structure */
	dev->thread = current;

	/* set signal mask to what we want to respond */
	siginitsetinv (&current->blocked,
		       sigmask(SIGKILL) | sigmask(SIGINT) | sigmask(SIGTERM));

	/* initialise wait queue head */
	init_waitqueue_head (&dev->queue);

	/* let others run */
	unlock_kernel();

	/* tell the creator that we are ready and let him run */
	up (&dev->sem);
}

static void leave_thread (kbd_dev_t *dev)
{
	/* lock the kernel, the exit will unlock it */
	dev->thread = NULL;

	/* notify the kill_thread() routine that we are terminating. */
	up (&dev->sem);
}
