/***********************************************************************
 *
 * (C) Copyright 2001, 2002
 * DENX Software Engineering
 * Wolfgang Denk, wd@denx.de
 * All rights reserved.
 *
 * This code is under GPL.
 *
 * This device driver uses the MPC8xx RISC timers to implement a
 * couple of PWM channels:
 *
 * - buzzer: this device operates a "horn" to emit a "beep" with
 *           adjustable length and frequency
 *
 * - lcdblc: this device controls the brightness of the LCD's
 *           backlight
 *
 ***********************************************************************/

#include <linux/kernel.h>
#include <linux/init.h>
#include <linux/module.h>
#include <linux/proc_fs.h>
#include <linux/spinlock.h>
#include <asm/uaccess.h>
#include <asm/io.h>
#include <asm/8xx_immap.h>
#include <asm/mpc8xx.h>
#include "commproc.h"
#include "pwm_8xx.h"

#undef	DEBUG

#ifdef	DEBUG
# define debugk(fmt,args...)	printk(fmt ,##args)
#else
# define debugk(fmt,args...)
#endif

/*
 * Deal with CONFIG_MODVERSIONS
 */
#if CONFIG_MODVERSIONS==1
/* # define MODVERSIONS */
# include <linux/modversions.h>
#endif

/*
 * For character devices
 */
#include <linux/fs.h>		/* character device definitions	*/
#include <linux/wrapper.h>	/* wrapper for compatibility with future versions */

#ifndef KERNEL_VERSION
# define KERNEL_VERSION(a,b,c)	((a)*65536+(b)*256+(c))
#endif

#ifndef	PWM_8xx_MAJOR
#define	PWM_8xx_MAJOR	115	/* Buzzer => Console speaker driver */
#endif	/* PWM_8xx_MAJOR */


/*
 * The name for our device, as it will appear in /proc/devices
 */
#define DEVICE_NAME "pwm_8xx"

/*
 * /proc filesystem definitions
 */
struct proc_dir_entry *proc_pwm_8xx;	/* root for pwm_8xx files	*/
struct proc_dir_entry *pwm_buzzer_f;	/* buzzer frequency		*/
struct proc_dir_entry *pwm_buzzer_d;	/* buzzer duration		*/
struct proc_dir_entry *pwm_lcd_blc;	/* LCD backlight control	*/

/*
 * prevent concurent access of the same device
 */
static int Device_Open = 0;

/*
 * Prototypes
 */
static	int	pwm_8xx_init (void) __init;
static	int	pwm_8xx_open (struct inode *, struct file *);
static	int	pwm_8xx_release (struct inode *, struct file *);
static	ssize_t	pwm_8xx_read  (struct file *, char *, size_t, loff_t *);
static	ssize_t	pwm_8xx_write (struct file *, const char *, size_t, loff_t *);
static	ssize_t	pwm_8xx_ioctl (struct inode *, struct file *,
			       unsigned int, unsigned long);
int init_module(void);
void cleanup_module(void);

static int pwm_buzzer_f_read_proc (char *, char **, off_t, int, int *, void *);
static int pwm_buzzer_d_read_proc (char *, char **, off_t, int, int *, void *);
static int pwm_lcd_blc_read_proc (char *, char **, off_t, int, int *, void *);
static int pwm_buzzer_f_write_proc (struct file *, const char *, unsigned long, void *);
static int pwm_buzzer_d_write_proc (struct file *, const char *, unsigned long, void *);
static int pwm_lcd_blc_write_proc (struct file *, const char *, unsigned long, void *);
static int pwm_8xx_write_proc (unsigned long, const char *, int *);

struct file_operations pwm_8xx_ops =
{
	owner:		THIS_MODULE,
	open:		pwm_8xx_open,
	release:	pwm_8xx_release,
	read:		pwm_8xx_read,
	write:		pwm_8xx_write,
	ioctl:		pwm_8xx_ioctl,
};


static volatile rt_pram_t *rtp;		/* RISC Time PRAM pointer	*/
static volatile immap_t *immr;		/* IMMR				*/


#define	BUZZER_BIT	0x00000100	/* Buzzer is on Port B.23	*/
#define	BLC_BIT		0x00000200	/* Backlight Ctrl is B.22	*/

/*
 * TIMEP=0 creates a tick every (0+1) * 1024 = 1024 System Clocks.
 * At a Clock of 66 MHz this equals 64.45 kHz or about 15.5 us.
 *
 * Note: higher values of TIMEP are only needed to generate
 * frequencies
 *
 *	freq < clock / 1024 / 65536 [ = 66e6 / 1024 / 65536 = 0.98 Hz]
 *
 * Thus we can use a fix value of TIMEP here, and need only
 * adjustment of PERIOD.
 *
 * PERIOD computes as follows:
 *
 *          clock   (TIMEP + 1)
 * period = ----- * -----------
 *           1024      freq
 *
 * For instance: clock = 66MHz, TIMEP = 0, freq = 1 kHz ==> period = 64
 */
#define MK_PERIOD(clk,t,f) ((((clk)/1024)*(t+1))/(f))

#define	CONFIG_TIMEP	0

MODULE_AUTHOR("Wolfgang Denk, wd@denx.de");
MODULE_DESCRIPTION("Buzzer Driver for LWMON Project");
MODULE_PARM(freq, "i");
MODULE_PARM_DESC(freq, "Frequency of Beep");
MODULE_PARM(duration, "i");
MODULE_PARM_DESC(duration, "Duration of Beep");

static int freq		= 3275;		/* Initialize with 3.3 kHz	*/
static int duration	= 500;		/* Duration in ms		*/
static int brightness	= 100;		/* brightness in %		*/


static void set_frequency (int frequency);
static void set_backlight (int brightness);
static void buzzer_off (ulong);

static struct timer_list buzzer_timer = {
	function:	buzzer_off,
	data:		0L,
};

static uint tm_base;

/*
 * Initialize the driver - Register the character device
 */
static int __init pwm_8xx_init (void)
{
	int rc;
	ulong	val;
	ushort	sval;

	debugk (__FUNCTION__ "() called\n");

	/* init CPM ptr */
	asm( "mfspr %0,638": "=r"(val) : );
	val &= 0xFFFF0000;
	immr = (volatile immap_t *)val;

	/*
	 * Register the character device
	 */
	rc = register_chrdev (PWM_8xx_MAJOR, DEVICE_NAME, &pwm_8xx_ops);

	/* Negative value on error */
	if (rc < 0) {
		printk (DEVICE_NAME ": init maj=%d failed rc=%d\n",
			PWM_8xx_MAJOR, rc);
		return (rc);
	}

	printk (KERN_INFO DEVICE_NAME ": $Revision: 1.1.1.1 $ wd@denx.de\n");

	debugk (DEVICE_NAME " registered: major = %d  freq = %d  duration = %d\n",
		PWM_8xx_MAJOR, freq, duration);

	if ( ((proc_pwm_8xx = proc_mkdir (DEVICE_NAME, NULL)) == NULL)	||
	     ((pwm_buzzer_f = create_proc_entry("buzzer_f",
					   S_IRUGO | S_IWUSR,
					   proc_pwm_8xx)) == NULL) 	||
	     ((pwm_buzzer_d = create_proc_entry("buzzer_d",
					   S_IRUGO | S_IWUSR,
					   proc_pwm_8xx)) == NULL) 	||
	     ((pwm_lcd_blc  = create_proc_entry("lcd_blc",
					   S_IRUGO | S_IWUSR,
					   proc_pwm_8xx)) == NULL) 	) {
		printk (DEVICE_NAME ": proc_register failed\n");
		return (-ENXIO);
	}

	pwm_buzzer_f->read_proc  = pwm_buzzer_f_read_proc;
	pwm_buzzer_f->write_proc = pwm_buzzer_f_write_proc;

	pwm_buzzer_d->read_proc  = pwm_buzzer_d_read_proc;
	pwm_buzzer_d->write_proc = pwm_buzzer_d_write_proc;

	pwm_lcd_blc->read_proc   = pwm_lcd_blc_read_proc;
	pwm_lcd_blc->write_proc  = pwm_lcd_blc_write_proc;

	/*
	 * Get pointer to Risc Timer Parameter RAM
	 */
	rtp = (rt_pram_t *)&cpmp->cp_dparam[PROFF_RTMR];

	/* (1)
	 * Set RISC Timer Period; don't set Timer Enable yet
	 */
	sval = (immr->im_cpm.cp_rccr & RCCR_TIME_MASK) | RCCR_TIMEP(CONFIG_TIMEP);
	debugk (DEVICE_NAME ": RCCR = 0x%04X\n", sval);
	immr->im_cpm.cp_rccr = sval;

	/* (2)
	 * We need two pair of timers, thus 2 * 2 * 4 bytes in DPRAM
	 */
	if ((tm_base = m8xx_cpm_dpalloc(2 * 2 * 4)) == CPM_DP_NOSPACE)
		panic (DEVICE_NAME "can't allocate RISC Timer Table");
	
	rtp->tm_base = (ushort)tm_base;

	debugk (DEVICE_NAME ": tm_base = %04X (%X)\n", rtp->tm_base, tm_base);
	/* (3) Clear TM_CNT */
	rtp->tm_cnt = 0;

	/* (4) Clear any pending events in RTER */
	immr->im_cpm.cp_rter = 0xFFFF;

	/* (5) No changes to RTMR - we don't use interrupts */
	/* (6) No changes to CIMR[RTT] - we don't use interrupts */

	/*
	 * Configure buzzer and backlight control ports
	 * We set the port pins to input mode, so the signal is off initally
	 */
	immr->im_cpm.cp_pbpar &= ~(BUZZER_BIT | BLC_BIT);
	immr->im_cpm.cp_pbodr &= ~(BUZZER_BIT | BLC_BIT);
	immr->im_cpm.cp_pbdat &= ~(BUZZER_BIT | BLC_BIT);
	immr->im_cpm.cp_pbdir &= ~(BUZZER_BIT | BLC_BIT);

	set_frequency (freq);		/* Starts RISC timers 0 + 1	*/

	set_backlight (brightness);	/* Starts RISC timers 2 + 3	*/

	/* (9)
	 * Start timers
	 */
	immr->im_cpm.cp_rccr |= RCCR_TIME;

	return (0);
}

/*
 * Stop timer, set new frequency, and start timer
 */
static void set_frequency (int frequency)
{
	bd_t	*bd;
	ushort	period;

	debugk (DEVICE_NAME ": set frequency=%d\n", frequency);

	bd = (bd_t *)__res;

	period = MK_PERIOD((bd->bi_intfreq*1000000),CONFIG_TIMEP,frequency);

	/* (7)
	 * Configure TM_CMD:
	 *
	 * We use timers 0 + 1 to create a PWM channel on Port B.23
	 * Duty cycle is (approx.) 50%
	 */
	rtp->tm_cmd  =	TM_CMD_NUM(0)	| TM_CMD_PERIOD(period/2 + 1) | \
			TM_CMD_PWM	| TM_CMD_VALID;

	/* (8) Issue SET TIMER command to CP */
	immr->im_cpm.cp_cpcr = mk_cr_cmd (CPM_CR_CH_TIMER, CPM_CR_SET_TIMER) | \
			       CPM_CR_FLG;
	while (immr->im_cpm.cp_cpcr & CPM_CR_FLG)
		;

	/* repeat for second timer */
	rtp->tm_cmd  =	TM_CMD_NUM(1)	| TM_CMD_PERIOD(period + 1) | \
			TM_CMD_RESTART	| TM_CMD_VALID;

	immr->im_cpm.cp_cpcr = mk_cr_cmd (CPM_CR_CH_TIMER, CPM_CR_SET_TIMER) | \
			       CPM_CR_FLG;
	while (immr->im_cpm.cp_cpcr & CPM_CR_FLG)
		;
}

/*
 * Switch buzzer off
 */
static void buzzer_off (ulong dummy)
{
	debugk (__FUNCTION__ "() called\n");

	/* disable output */
	immr->im_cpm.cp_pbdir &= ~(BUZZER_BIT);
}

/*
 * Set LCD Backlight brightness level (0 ... 100%)
 */
static void set_backlight (int brightness)
{
	bd_t	*bd;
	ushort	period, active;

	debugk (__FUNCTION__ "(%d) called\n", brightness);

	bd = (bd_t *)__res;

	/*
	 * Run LCD Backlight Controll at about 100 Hz,
	 * but round period to a multiple of 100
	 */
	period = MK_PERIOD((bd->bi_intfreq*1000000),CONFIG_TIMEP,100);
	period = ((period + 99) / 100) * 100;

	/* (7)
	 * Configure TM_CMD:
	 *
	 * We use timers 2 + 3 to create a PWM channel on Port B.22
	 */
	active = ((int) period * brightness) / 100;

	debugk ("LCD PWM: %d/%d = %d\n", active, period, active / period);

	rtp->tm_cmd  =	TM_CMD_NUM(2)	| TM_CMD_PERIOD(active + 1) | \
			TM_CMD_PWM	| TM_CMD_VALID;

	/* (8) Issue SET TIMER command to CP */
	immr->im_cpm.cp_cpcr = mk_cr_cmd (CPM_CR_CH_TIMER, CPM_CR_SET_TIMER) | \
			       CPM_CR_FLG;
	while (immr->im_cpm.cp_cpcr & CPM_CR_FLG)
		;

	/* repeat for second timer */
	rtp->tm_cmd  =	TM_CMD_NUM(3)	| TM_CMD_PERIOD(period + 1) | \
			TM_CMD_RESTART	| TM_CMD_VALID;

	immr->im_cpm.cp_cpcr = mk_cr_cmd (CPM_CR_CH_TIMER, CPM_CR_SET_TIMER) | \
			       CPM_CR_FLG;
	while (immr->im_cpm.cp_cpcr & CPM_CR_FLG)
		;

	/* Now make sure the port pin is an output */
	immr->im_cpm.cp_pbdir |= BLC_BIT;
}

/*
 * called whenever a process attempts to open the device
 */
static int pwm_8xx_open (struct inode *inode, struct file *file)
{
	debugk (__FUNCTION__ "() called\n");

	/*
	 * Get major / minor numbers when needed
	 */
	debugk (DEVICE_NAME ": Device: major=%d minor=%d\n",
		inode->i_rdev >> 8, inode->i_rdev & 0xFF);

	if (!immr) {
		debugk (DEVICE_NAME ": init failed: immr==NULL\n");
		return (-ENXIO);
	}

	/*
	 * exclusive open only
	 */
	if (Device_Open) {
		return (-EBUSY);
	}

	/*
	 * Be careful here on SMP kernels!
	 */
	Device_Open++;

	/*
	 * Make sure that the module isn't removed while
	 * the file is open by incrementing the usage count
	 */
	MOD_INC_USE_COUNT;

	return (0);
}


/*
 * Called when a process closes the device.
 */
static int pwm_8xx_release (struct inode *inode, struct file *file)
{
	debugk (__FUNCTION__ "() called\n");

	Device_Open--;

	MOD_DEC_USE_COUNT;

	return (0);
}


/*
 * read entry point:
 */

static ssize_t pwm_8xx_read (struct file *file,
			  char *buffer,		/* The buffer to fill with data	*/
			  size_t length,	/* The length of the buffer	*/
			  loff_t * offset)	/* Our offset in the file	*/
{
	debugk (__FUNCTION__ "() called: len=%d\n", length);

	return (0);		/* read() normally returns length	*/
}


/*
 * write entry point:
 */
static ssize_t pwm_8xx_write (struct file *file,
			     const char *buffer,	/* buffer		*/
			     size_t length,		/* length of buffer	*/
			     loff_t * offset)		/* offset in the file	*/
{
	int exp;
	unsigned long flags;

	debugk (__FUNCTION__ "() called: len=%d\n", length);

	/* enable output, starting with low */
	immr->im_cpm.cp_pbdat &= ~(BUZZER_BIT);
	immr->im_cpm.cp_pbdir |=   BUZZER_BIT;

	if ((exp = (duration * HZ) / 1000) == 0)
		exp = 1;

	debugk ("Expires: %d [%ld / %ld]\n", exp, jiffies, buzzer_timer.expires);

	save_flags(flags);
	cli();

	/* Just in case a previous timer is still running: delete it */
	del_timer(&buzzer_timer);

	/* Set up new timeout */
	init_timer(&buzzer_timer);
	buzzer_timer.expires = jiffies + exp;
	add_timer(&buzzer_timer);

	restore_flags(flags);

	return (length);
}


/*
 * ioctl entry point:
 */
static ssize_t pwm_8xx_ioctl(struct inode *inode, struct file *file,
			  unsigned int cmd, unsigned long arg)
{
	buzzer_param_t param;
	int val;

	debugk (__FUNCTION__ "() called: cmd=%X\n", cmd);

	switch (cmd) {
	case BUZZER_GET:
		debugk ("IOCTL buzzer get: f=%d d=%d\n", freq, duration);
		param.freq	= freq;
		param.duration	= duration;
		return (copy_to_user((void *)arg, &param, sizeof (param)));
		break;
	case BUZZER_SET:
		if (copy_from_user(&param, (void *)arg, sizeof (param))) {
			return (-EFAULT);
		}
		freq	 = param.freq;
		duration = param.duration;
		debugk ("IOCTL buzzer set: f=%d d=%d\n", freq, duration);
		set_frequency (freq);
		break;
	case LCDBLC_GET:
		val = brightness;
		debugk ("IOCTL LCD BLC get: blc=%d%%\n", val);
		
		return (copy_to_user((void *)arg, &val, sizeof (val)));
		break;
	case LCDBLC_SET:
		if (copy_from_user(&val, (void *)arg, sizeof (val))) {
			return (-EFAULT);
		}
		if ((val < 0) || (val > 100))
			return (-EDOM);

		debugk ("IOCTL LCD BLC set: blc=%d%%\n", val);
		brightness = val;
		set_backlight (brightness);
		break;
	default:
		debugk ("IOCTL pwm_8xx: invalid command\n");
		return (-EINVAL);
	}
	return (0);
}


/******************************
 **** Module Declarations *****
 **************************** */

module_init (pwm_8xx_init);

#ifdef MODULE
/*
 * Cleanup - unregister the driver
 */
static void pwm_8xx_cleanup (void)
{
	int rc;

	debugk (__FUNCTION__ "() called\n");

	/* cancel any timer that might be running */
	del_timer(&buzzer_timer);

	immr->im_cpm.cp_rccr &= ~RCCR_TIME;

	/*
	 * Make sure buzzer output is off,
	 * LCD baclight control is off (max brightness)
	 */
	immr->im_cpm.cp_pbdir &= ~(BUZZER_BIT | BLC_BIT);

	/*
	 * Unregister the device
	 */
	rc = unregister_chrdev (PWM_8xx_MAJOR, DEVICE_NAME);

	/*
	 * If there's an error, report it
	 */
	if (rc < 0) {
		printk (DEVICE_NAME "unregister_chrdev: error %d\n", rc);
	}

	m8xx_cpm_dpfree (tm_base);

	remove_proc_entry ("lcd_blc",  proc_pwm_8xx);
	remove_proc_entry ("buzzer_d", proc_pwm_8xx);
	remove_proc_entry ("buzzer_f", proc_pwm_8xx);
	remove_proc_entry (DEVICE_NAME, NULL);
}

module_exit (pwm_8xx_cleanup);

#endif	/* MODULE */

/* ==================================================================== */

static int pwm_buzzer_f_read_proc (char *page, char **start, off_t off,
				   int count, int *eof, void *data)
{
	debugk (__FUNCTION__ "freq=%d\n", freq);
	return (sprintf (page, "%d\n", freq));
}

static int pwm_buzzer_d_read_proc (char *page, char **start, off_t off,
				   int count, int *eof, void *data)
{
	debugk (__FUNCTION__ "duration=%d\n", duration);
	return (sprintf (page, "%d\n", duration));
}

static int pwm_lcd_blc_read_proc (char *page, char **start, off_t off,
				   int count, int *eof, void *data)
{
	debugk (__FUNCTION__ "brightness=%d\n", brightness);
	return (sprintf (page, "%d\n", brightness));
}

static int pwm_buzzer_f_write_proc (struct file *file, const char *buffer,
				    unsigned long count, void *data)
{
	int len = pwm_8xx_write_proc (count, buffer, &freq);
	set_frequency (freq);

	return (len);
}

static int pwm_buzzer_d_write_proc (struct file *file, const char *buffer,
				    unsigned long count, void *data)
{
	return (pwm_8xx_write_proc (count, buffer, &duration));
}

static int pwm_lcd_blc_write_proc (struct file *file, const char *buffer,
				    unsigned long count, void *data)
{
	int len = pwm_8xx_write_proc (count, buffer, &brightness);
	if (brightness < 0)
		brightness = 0;
	if (brightness > 100)
		brightness = 100;
	set_backlight (brightness);

	return (len);
}

static int pwm_8xx_write_proc (unsigned long count, const char *buffer,
			       int *parm)
{
	char tmp[512];
	int len;

	if (!count)
		return (count);
	
	len = sizeof(tmp) - 1;
	if (count < len)
		len = count;
	
	copy_from_user (tmp, buffer, len);

	tmp[len] = '\0';

	debugk (__FUNCTION__ "buffer: \"%s\"\n", tmp);

	*parm = simple_strtoul ((const char *)&tmp, NULL, 0);

	return (len);
}

/* ==================================================================== */
