/*
 * arch/ppc/kernel/pcippc2_pci.c
 * 
 * PCI support for Gespac PCIPPC boards.
 *
 * (C) 2002 DENX Software Engineering, Wolfgang Denk, wd@denx.de
 *
 * Based on the K2 board support from HHL.
 *
 * Author: Matt Porter <mporter@mvista.com>
 *
 * Copyright 2001 MontaVista Software Inc.
 *
 * This program is free software; you can redistribute  it and/or modify it
 * under  the terms of  the GNU General  Public License as published by the
 * Free Software Foundation;  either version 2 of the  License, or (at your
 * option) any later version.
 */

#include <linux/kernel.h>
#include <linux/init.h>
#include <linux/pci.h>
#include <linux/malloc.h>

#include <asm/byteorder.h>
#include <asm/io.h>
#include <asm/uaccess.h>
#include <asm/machdep.h>
#include <asm/pci-bridge.h>
#include <asm/pcippc2.h>

#include "pci.h"
#include "pci_auto.h"
#include "cpc710.h"

#undef DEBUG
#ifdef DEBUG
#define DBG(x...) printk(x)
#else
#define DBG(x...)
#endif /* DEBUG */

/* Interrupt routes from cPCI and local buses */
static unsigned char PCIPPC_CPCI_IRQ_routes[4] = {PCI_INTA, PCI_INTB, PCI_INTC, PCI_INTD};
static unsigned char PCIPPC_LOCAL_IRQ_routes[6] = {0xff, 0xff, ETHERNET1_INT, ETHERNET2_INT, FIREWIRE_INT, SCSI_INT1};
static unsigned char PCIPPC_PMC_IRQ_routes[4] = {LPCI_INTA, LPCI_INTB, LPCI_INTC, LPCI_INTD};

static inline unsigned char bridge_swizzle(unsigned char pin,
		unsigned char idsel) 
{
	return (((pin-1) + idsel) % 4) + 1;
}

void
pcippc2_route_cpci_int(struct pci_dev *dev)
{
	struct pci_bus  *pbus;          /* Parent bus structure pointer */
	struct pci_dev	*tdev = dev;	/* Temporary pci_dev pointer */
	unsigned int    devnum;         /* Accumulated device number */
	unsigned char   intpin;         /* PCI interrupt pin */
	struct pci_controller *hose;	/* PCI controller */

	/* Check for valid PCI dev pointer */
	if (dev == NULL) return;

	/* Get a pointer to our pci_controller descriptor */
	hose = pci_bus_to_hose(dev->bus->number);

	/* Initialize bridge IDSEL variable */
	devnum = PCI_SLOT(dev->devfn);

	/* Read the interrupt pin of the device */
	pcibios_read_config_byte(dev->bus->number, dev->devfn,
			PCI_INTERRUPT_PIN, &intpin);

	/* If device doesn't request an interrupt, return */
	if ( (intpin < 1) || (intpin > 4) )
		return;

	/* Only loop if we are on a subordinate bus  */
	if (dev->bus->number != hose->first_busno)
	{
		/*
		 * Walk up to the top bus, adjusting the interrupt pin for the
		 * standard PCI bus swizzle.
		 */ 
		do
		{
			intpin = bridge_swizzle(intpin, devnum);
			pbus = tdev->bus;        /* up one level */
			tdev = pbus->self;
			devnum = PCI_SLOT(tdev->devfn);
		} while(tdev->bus->number != hose->first_busno);
	}

	/*
	 * devnum is reduced by 6 because the CPC710
	 * addresses IDSELs differently than the PCI
	 * spec suggests.
	 */
	intpin = bridge_swizzle(intpin, devnum-6);

	/*
	 * For the second PCI hose, use the swizzled interrupt pin
	 * value to map to a CPCI interrupt input on the PIC
	 */
	dev->irq = PCIPPC_CPCI_IRQ_routes[intpin-1];

	/* Write calculated interrupt value to header and device list */
	pci_write_config_byte(dev, PCI_INTERRUPT_LINE, dev->irq);
}

void
pcippc2_route_local_int(struct pci_dev *dev)
{
	unsigned int    devnum;         /* Accumulated device number */
	unsigned char   intpin;         /* PCI interrupt pin */

	/* Check for valid PCI dev pointer */
	if (dev == NULL) return;

	/* Initialize bridge IDSEL variable */
	devnum = PCI_SLOT(dev->devfn);

	/* If device doesn't request an interrupt, return */
	if ( (devnum < 1) || (devnum > PCIPPC_PMC_DEVNO) )
		return;

	/* Read the interrupt pin of the device */
	pcibios_read_config_byte(dev->bus->number, dev->devfn,
			PCI_INTERRUPT_PIN, &intpin);

	if ( devnum == PCIPPC_PMC_DEVNO )
  	dev->irq = PCIPPC_PMC_IRQ_routes[intpin];
	else
		dev->irq = PCIPPC_LOCAL_IRQ_routes[devnum];

	/* Write calculated interrupt value to header and device list */
	pci_write_config_byte(dev, PCI_INTERRUPT_LINE, dev->irq);
}

void pcippc2_pcibios_fixup(void)
{ 
	struct pci_dev *dev;

	pci_for_each_dev(dev)
	{
		if (pci_bus_to_hose(dev->bus->number)->first_busno)
			pcippc2_route_cpci_int(dev);
		else
			pcippc2_route_local_int(dev);
		}
}

void pcippc2_pcibios_fixup_resources(struct pci_dev *dev)
{
	int i;

	if ((dev->vendor == PCI_VENDOR_ID_IBM) &&
			(dev->device == PCI_DEVICE_ID_IBM_CPC710_PCI64))
	{	
		DBG("Fixup CPC710 resources\n");
		for (i=0; i<DEVICE_COUNT_RESOURCE; i++)
		{
			dev->resource[i].start = 0;
			dev->resource[i].end = 0;
		}
	}
}

void pcippc2_find_bridges(void)
{
	struct pci_controller *hose_a, *hose_b;

	/* Setup PCI32 hose */
	hose_a = pcibios_alloc_controller();
	if (!hose_a)
		return;

	hose_a->first_busno = 0;
	hose_a->last_busno = 0xff;
	hose_a->pci_mem_offset = PCIPPC_PCI32_MEM_BASE - PCIPPC_PCI32_LOWER_MEM;
	hose_a->mem_resources[0].end = 0xffffffff;
	hose_a->io_space.start = PCIPPC_PCI32_LOWER_IO;
	hose_a->io_space.end = PCIPPC_PCI32_UPPER_IO;
	hose_a->mem_space.start = PCIPPC_PCI32_LOWER_MEM;
	hose_a->mem_space.end = PCIPPC_PCI32_UPPER_MEM;
	hose_a->io_base_virt = (void *)(PCIPPC_PCI32_IO_BASE - PCIPPC_PCI32_LOWER_IO);

	setup_indirect_pci(hose_a, PCIPPC_PCI32_CONFIG_ADDR, PCIPPC_PCI32_CONFIG_DATA);

	/* Initialize PCI32 bus registers */
	early_write_config_byte(hose_a,
			hose_a->first_busno,
			PCI_DEVFN(0, 0),
			CPC710_BUS_NUMBER,
			hose_a->first_busno);
	early_write_config_byte(hose_a,
			hose_a->first_busno,
			PCI_DEVFN(0, 0),
			CPC710_SUB_BUS_NUMBER,
			hose_a->last_busno);

	hose_a->last_busno = pciauto_bus_scan(hose_a, hose_a->first_busno);

	/* Write out correct max subordinate bus number for hose A */
	early_write_config_byte(hose_a,
			hose_a->first_busno,
			PCI_DEVFN(0, 0),
			CPC710_SUB_BUS_NUMBER,
			hose_a->last_busno);

	/* setup PCI64 hose */
	if ( 1 )
	{
		/* Setup PCI64 hose */
		hose_b = pcibios_alloc_controller();
		if (!hose_b)
			return;

		hose_b->first_busno = hose_a->last_busno + 1;
		hose_b->last_busno = 0xff;
		/* Reminder: quit changing the following, it is correct. */
		hose_b->pci_mem_offset = PCIPPC_PCI64_MEM_BASE - PCIPPC_PCI64_LOWER_MEM;
		hose_b->mem_resources[0].end = 0xffffffff;
		hose_b->io_space.start = PCIPPC_PCI64_LOWER_IO;
		hose_b->io_space.end = PCIPPC_PCI64_UPPER_IO;
		hose_b->mem_space.start = PCIPPC_PCI64_LOWER_MEM;
		hose_b->mem_space.end = PCIPPC_PCI64_UPPER_MEM;
		hose_b->io_base_virt = (void *)(PCIPPC_PCI64_IO_BASE - PCIPPC_PCI64_LOWER_IO);

		setup_indirect_pci(hose_b,
				PCIPPC_PCI64_CONFIG_ADDR,
				PCIPPC_PCI64_CONFIG_DATA);

		/* Initialize PCI64 bus registers */
		early_write_config_byte(hose_b,
				0,
				PCI_DEVFN(0, 0),
				CPC710_SUB_BUS_NUMBER,
				0xff);

		early_write_config_byte(hose_b,
				0,
				PCI_DEVFN(0, 0),
				CPC710_BUS_NUMBER,
				hose_b->first_busno);

		hose_b->last_busno = pciauto_bus_scan(hose_b,
				hose_b->first_busno);

		/* Write out correct max subordinate bus number for hose B */
		early_write_config_byte(hose_b,
				hose_b->first_busno,
				PCI_DEVFN(0, 0),
				CPC710_SUB_BUS_NUMBER,
				hose_b->last_busno);
	}

	ppc_md.pcibios_fixup = pcippc2_pcibios_fixup;
}
