/***********************************************************************
 *
 * (C) Copyright 2002
 * DENX Software Engineering
 * Wolfgang Denk, wd@denx.de
 * All rights reserved.
 *
 * This driver provides access to the Gespac PCIPPC RTC
 *
 ***********************************************************************/

/*
 * Standard in kernel modules
 */
#include <linux/kernel.h>
#include <linux/delay.h>
#include <linux/module.h>
#include <linux/miscdevice.h>
#include <linux/devfs_fs_kernel.h>
#include <linux/rtc.h>
#include <linux/smp_lock.h>
#include <linux/version.h>
#include <asm/types.h>
#include <asm/uaccess.h>
#include <asm/init.h>
#include <asm/io.h>
#include <asm/time.h>
#include <asm/machdep.h>		/* ppc_md */
#include <asm/mk48t59.h>

#ifndef KERNEL_VERSION
# define KERNEL_VERSION(a,b,c)	(((a) << 16) + ((b) << 8) + (c))
#endif

#if (!defined(LINUX_VERSION_CODE) || LINUX_VERSION_CODE < KERNEL_VERSION(2,2,0))
#error "You need to use at least Linux kernel version 2.2.0"
#endif

/***************************************************************************
 * Global stuff
 ***************************************************************************
 */
/* #define	DEBUG	1	//	*/

/*
 * For character devices
 */
#include <linux/fs.h>		/* character device definitions	*/
#include <linux/wrapper.h>	/* wrapper for compatibility with future versions */


/*
 * Device Declarations
 */

/*
 * The name for our device, as it will appear in /proc/devices
 */
#define DEVICE_NAME	"rtc"
#define	RTC_VERSION	"$Revision: 1.1.1.1 $ wd@denx.de"

/***************************************************************************
 * Local stuff
 ***************************************************************************
 */

#ifdef	DEBUG
# define debugk(fmt,args...)	printk(fmt ,##args)
# define DEBUG_OPTARG(arg)	arg,
#else
# define debugk(fmt,args...)
# define DEBUG_OPTARG(arg)
#endif

/*
 * Prototypes
 */
int mk48t59_rtc_init (void);
static int mk48t59_rtc_open	(struct inode *, struct file *	);
static int mk48t59_rtc_release	(struct inode *, struct file *	);
static int mk48t59_rtc_ioctl	(struct inode *, struct file *,
				 unsigned int, unsigned long	);
static unsigned long	mk48t59_get_rtc_time(void);
static int		mk48t59_set_rtc_time(unsigned long now);
static void		mk48t59_reset_rtc(void);

int init_module(void);
void cleanup_module(void);

/* avoid multiple access */
static char rtc_status = 0;

/* save/restore old machdep pointers */
int		(*save_set_rtc_time)(unsigned long);
unsigned long	(*save_get_rtc_time)(void);

/***************************************************************************
 * Driver interface
 ***************************************************************************
 */
#if LINUX_VERSION_CODE < KERNEL_VERSION(2,3,0)

static struct file_operations mk48t59_rtc_fops = {
	NULL,			/* lseek	*/
	NULL,			/* read		*/
	NULL,			/* write	*/
	NULL,			/* readdir	*/
	NULL,			/* select	*/
	mk48t59_rtc_ioctl,	/* ioctl	*/
	NULL,			/* mmap		*/
	mk48t59_rtc_open,		/* open		*/
	NULL,			/* flush	*/
	mk48t59_rtc_release	/* close	*/
};

#else	/* Linux kernel >= 2.3.x */

static struct file_operations mk48t59_rtc_fops = {
	owner:			THIS_MODULE,
	open:			mk48t59_rtc_open,
	release:		mk48t59_rtc_release,
	ioctl: 			mk48t59_rtc_ioctl,
};
#endif

/*
 * We sponge a minor off of the misc major. No need slurping
 * up another valuable major dev number for this.
 */
static struct miscdevice rtc_dev = {
	RTC_MINOR,
	DEVICE_NAME,
	&mk48t59_rtc_fops,
};


/***************************************************************************
 *
 * Initialize the driver - Register the character device
 */
int __init mk48t59_rtc_init (void)
{
	unsigned long now, flags;
	extern time_t last_rtc_update;
	extern rwlock_t xtime_lock;

	printk (KERN_INFO "MK48T59 Real-Time Clock Driver " RTC_VERSION "\n");

	misc_register(&rtc_dev);

	/* Switching kernel RTC pointers */
	debugk ("RTC switching kernel pointers\n");

	save_set_rtc_time   = ppc_md.set_rtc_time;
	save_get_rtc_time   = ppc_md.get_rtc_time;

	ppc_md.set_rtc_time = mk48t59_set_rtc_time;
	ppc_md.get_rtc_time = mk48t59_get_rtc_time;

	/*
	 * Set system time
	 * Code copied from arch/ppc/kernel/time.c
	 */
	write_lock_irqsave(&xtime_lock, flags);

	now = mk48t59_get_rtc_time();

	debugk ("Set System Time from RTC Time: %lu\n", now);

	xtime.tv_usec = 0;
	xtime.tv_sec  = now;

	last_rtc_update = now - 658;

	time_adjust   = 0;		/* stop active adjtime() */
	time_status  |= STA_UNSYNC;
	time_state    = TIME_ERROR;
	time_maxerror = NTP_PHASE_LIMIT;
	time_esterror = NTP_PHASE_LIMIT;

	write_unlock_irqrestore(&xtime_lock, flags);

	debugk ("RTC installed, minor %d\n", RTC_MINOR);

	/*
	 * Reset any error conditions, alarms, etc.
	 */
	mk48t59_reset_rtc ();

	return 0;
}

/***************************************************************************
 *
 * called whenever a process attempts to open the device
 */
static int mk48t59_rtc_open (struct inode *inode, struct file *file)
{
	lock_kernel();
	if (rtc_status) {
		unlock_kernel();
		return -EBUSY;
	}

	rtc_status = 1;
	unlock_kernel();

	MOD_INC_USE_COUNT;

	return 0;
}



/***************************************************************************
 *
 * Called when a process closes the device.
 */
static int mk48t59_rtc_release (struct inode *inode, struct file *file)
{
	lock_kernel();
	rtc_status = 0;
	unlock_kernel();

	MOD_DEC_USE_COUNT;

	return 0;
}


/***************************************************************************
 *
 * ioctl entry point:
 */
static int mk48t59_rtc_ioctl (struct inode *inode, struct file *file,
	unsigned int cmd, unsigned long arg)
{
	struct rtc_time tm;
	int leap_yr;
	static const unsigned char days_in_month[] =
		{ 0, 31, 28, 31, 30, 31, 30, 31, 31, 30, 31, 30, 31 };

	switch (cmd) {
	case RTC_RD_TIME:	/* Read the time/date from RTC	*/

		to_tm (mk48t59_get_rtc_time(), &tm);

		/* convert from rtc_time to struct tm conventions */
		tm.tm_year -= 1900;
		tm.tm_mon  -= 1;

		debugk ("RTC_RD_TIME: %4d-%02d-%02d-%02d:%02d:%02d\n",
			tm.tm_year+1900, tm.tm_mon+1, tm.tm_mday,
			tm.tm_hour, tm.tm_min, tm.tm_sec);

		return (copy_to_user((void *)arg, &tm, sizeof(tm))) ? -EFAULT : 0;

	case RTC_SET_TIME:	/* Set the time/date in the RTC	*/

		if (!capable(CAP_SYS_TIME))
			return -EACCES;

		if (copy_from_user(&tm, (struct rtc_time *)arg, sizeof(tm)))
			return -EFAULT;

		/* convert from struct tm to rtc_time conventions */
		tm.tm_year += 1900;
		tm.tm_mon  += 1;

		/* simple but OK for all dates in time_t range */
		leap_yr = ((tm.tm_year % 4) == 0);

		if ((tm.tm_year < 1970) ||
		    (tm.tm_mon  > 12)	||
		    (tm.tm_mday == 0)	||
		    (tm.tm_mday >
		    	(days_in_month[tm.tm_mon] + ((tm.tm_mon==2)&&leap_yr)))	||
		    (tm.tm_hour >= 24)	||
		    (tm.tm_min  >= 60)	||
		    (tm.tm_sec  >= 60)	) {

			return (-EINVAL);
		}

		debugk ("RTC_SET_TIME: %4d-%02d-%02d-%02d:%02d:%02d\n",
			tm.tm_year, tm.tm_mon, tm.tm_mday,
			tm.tm_hour, tm.tm_min, tm.tm_sec);

		mk48t59_set_rtc_time (mktime(tm.tm_year, tm.tm_mon, tm.tm_mday,
					   tm.tm_hour, tm.tm_min, tm.tm_sec ) );

		return (0);

	default:
		return (-EINVAL);
	}
}


/***************************************************************************
 * Hardware related stuff
 ***************************************************************************
 */

/***************************************************************************
 *
 * set RTC time:
 */
int mk48t59_set_rtc_time(unsigned long nowtime)
{
	unsigned char save_control;
	struct rtc_time tm;


	to_tm(nowtime, &tm);

	debugk ("Set RTC [dec] year=%d mon=%d day=%d hour=%d min=%d sec=%d\n",
		tm.tm_year, tm.tm_mon, tm.tm_mday,
		tm.tm_hour, tm.tm_min, tm.tm_sec);

	/* tell the clock it's being written */
	save_control = ppc_md.nvram_read_val(MK48T59_RTC_CONTROLA);
	
	ppc_md.nvram_write_val(MK48T59_RTC_CONTROLA,
			     (save_control | MK48T59_RTC_CA_WRITE));

        tm.tm_year = (tm.tm_year - 1900) % 100;
	BIN_TO_BCD(tm.tm_sec);
	BIN_TO_BCD(tm.tm_min);
	BIN_TO_BCD(tm.tm_hour);
	BIN_TO_BCD(tm.tm_mon);
	BIN_TO_BCD(tm.tm_mday);
	BIN_TO_BCD(tm.tm_year);

	debugk ("Set RTC [bcd] year=%X mon=%X day=%X "
		"hour=%X min=%X sec=%X wday=%X\n",
		tm.tm_year, tm.tm_mon, tm.tm_mday, tm.tm_hour, 
		tm.tm_min, tm.tm_sec, tm.tm_wday);

	ppc_md.nvram_write_val(MK48T59_RTC_SECONDS,      tm.tm_sec);
	ppc_md.nvram_write_val(MK48T59_RTC_MINUTES,      tm.tm_min);
	ppc_md.nvram_write_val(MK48T59_RTC_HOURS,        tm.tm_hour);
	ppc_md.nvram_write_val(MK48T59_RTC_MONTH,        tm.tm_mon);
	ppc_md.nvram_write_val(MK48T59_RTC_DAY_OF_MONTH, tm.tm_mday);
	ppc_md.nvram_write_val(MK48T59_RTC_YEAR,         tm.tm_year);
	
	/* Turn off the write bit. */
	ppc_md.nvram_write_val(MK48T59_RTC_CONTROLA, save_control);

	return 0;
}

/***************************************************************************
 *
 * get RTC time:
 */
unsigned long mk48t59_get_rtc_time(void)
{
	unsigned char save_control;
	unsigned int year, mon, day, hour, min, sec;

	/* Simple: freeze the clock, read it and allow updates again */
	save_control = ppc_md.nvram_read_val(MK48T59_RTC_CONTROLA);
	save_control &= ~MK48T59_RTC_CA_READ;
	ppc_md.nvram_write_val(MK48T59_RTC_CONTROLA, save_control);

	/* Set the register to read the value. */
	ppc_md.nvram_write_val(MK48T59_RTC_CONTROLA,
			     (save_control | MK48T59_RTC_CA_READ));

	sec = ppc_md.nvram_read_val(MK48T59_RTC_SECONDS);
	min = ppc_md.nvram_read_val(MK48T59_RTC_MINUTES);
	hour = ppc_md.nvram_read_val(MK48T59_RTC_HOURS);
	day = ppc_md.nvram_read_val(MK48T59_RTC_DAY_OF_MONTH);
	mon = ppc_md.nvram_read_val(MK48T59_RTC_MONTH);
	year = ppc_md.nvram_read_val(MK48T59_RTC_YEAR);

	debugk ("Get RTC year: %02x mon/cent: %02x mday: %02x wday: %02x "
		"hr: %02x min: %02x sec: %02x\n",
		year, mon_cent, mday, wday,
		hour, min, sec );

	/* Let the time values change again. */
	ppc_md.nvram_write_val(MK48T59_RTC_CONTROLA, save_control);

	BCD_TO_BIN(sec);
	BCD_TO_BIN(min);
	BCD_TO_BIN(hour);
	BCD_TO_BIN(day);
	BCD_TO_BIN(mon);
	BCD_TO_BIN(year);

	year = year + 1900;
	if (year < 1970) {
		year += 100;
	}

	debugk ("get rtc [dec] year=%d mon=%d day=%d hour=%d min=%d sec=%d\n",
		year, mon, day, hour, min, sec);

	return mktime(year, mon, day, hour, min, sec);
}

/***************************************************************************
 *
 * set RTC time:
 */
static void mk48t59_reset_rtc(void) {
	unsigned char tmp;

	tmp = ppc_md.nvram_read_val(MK48T59_RTC_CONTROLB);
	if (tmp & MK48T59_RTC_CB_STOP) {
		printk("Warning: RTC was stopped, date will be wrong.\n");
		ppc_md.nvram_write_val(MK48T59_RTC_CONTROLB, 
				       tmp & ~MK48T59_RTC_CB_STOP);
		/* Low frequency crystal oscillators may take a very long
		 * time to startup and stabilize. For now just ignore the
		 * the issue, but attempting to calibrate the decrementer
		 * from the RTC just after this wakeup is likely to be very 
		 * inaccurate. Firmware should not allow to load
		 * the OS with the clock stopped anyway...
		 */
	}
	/* Ensure that the clock registers are updated */
	tmp = ppc_md.nvram_read_val(MK48T59_RTC_CONTROLA);
	tmp &= ~(MK48T59_RTC_CA_READ | MK48T59_RTC_CA_WRITE);
	ppc_md.nvram_write_val(MK48T59_RTC_CONTROLA, tmp);
}

/***************************************************************************
 * Module Declarations
 ***************************************************************************
 */
#ifdef MODULE
/*
 * Initialize the module
 */
int init_module (void)
{
	mk48t59_rtc_init();

	return (0);
}

/*
 * Cleanup - unregister the appropriate file from /proc
 */
void cleanup_module (void)
{
	ppc_md.set_rtc_time = save_set_rtc_time;
	ppc_md.get_rtc_time = save_get_rtc_time;

	misc_deregister(&rtc_dev);

	debugk ("MK48T59 Real-Time Clock Driver de-registered\n");
}

#endif	/* MODULE */

