/*
 * (C) Copyright 2000
 * Jrg Haider, SIEMENS AG
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License as
 * published by the Free Software Foundation; either version 2 of
 * the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston,
 * MA 02111-1307 USA
 *
 * (C) 2001 Wolfgang Denk, wd@denx.de -- Cleanup, Modifications for 2.4 kernels
 * (C) 2001 Wolfgang Denk, wd@denx.de -- Adaption for MAX706TESA Watchdog (LWMON)
 * (C) 2001 Steven Hein,  ssh@sgi.com -- Added timeout configuration option
 * (C) 2001 Robert Enzinger, Robert.Enzinger.extern@icn.siemens.de,
 *		-- added ioctl() for dynamic configuration
 */

#include <linux/kernel.h>
#include <linux/module.h>
#include <linux/fs.h>			/* for character devices	*/
#include <linux/miscdevice.h>		/* driver is a misc device	*/
#include <linux/version.h>
#include <linux/init.h>			/* for __initfunc		*/
#include <asm/wdt_mpc8xx.h>
#include <asm/uaccess.h>		/* for put_user			*/
#include <asm/8xx_immap.h>
#include <asm/mpc8xx.h>

static volatile immap_t *immr = NULL;	/* pointer to register-structure	*/

/* #define DEBUG		//	*/
#define CONFIG_WDT_MPC8XX_TIMEOUT_OPEN_ONLY

#ifdef DEBUG
# define debugk(fmt,args...) printk(fmt ,##args)
#else
# define debugk(fmt,args...)
#endif

#define	WDT_VERSION	"0.5"	/* version of this device driver		*/

#ifdef	CONFIG_WDT_MPC8XX_TIMEOUT
#define TIMEOUT_VALUE CONFIG_WDT_MPC8XX_TIMEOUT	/* configurable timeout		*/
#else
#define TIMEOUT_VALUE 300	/* reset after five minutes = 300 seconds	*/
#endif

#ifdef CONFIG_WDT_MPC8XX_TIMEOUT_OPEN_ONLY
static int timeout_open_only = 1;
#else
static int timeout_open_only = 0;
#endif

struct timer_list wd_timer;	/* structure for timer administration		*/
static unsigned long wdt_mpc8xx_dummy;

/*
 * Watchdog active? When disabled, it will get re-triggered
 * automatically without timeout, so it appears to be switched off
 * although actually it is still running.
 */
static int enabled = 1;

/*
 * The shortest watchdog period of all boards is (so far) approx. 1 sec,
 * thus re-trigger watchdog by default every 500 ms = HZ / 2,
 * except for the LWMON board, which needs 100ms = HZ / 10.
 * The default can be overwritten using a boot argument.
 *
 * The external interface is in seconds, but internal all calculation
 * is done in jiffies.
 */
#ifdef	CONFIG_LWMON
# define TRIGGER_PERIOD   ( HZ / 10 )
#else
# define TRIGGER_PERIOD   ( HZ / 2 )
#endif

static unsigned long timer_count = TIMEOUT_VALUE * HZ;	/* remaining time	*/
static unsigned long timer_period= TRIGGER_PERIOD;	/* period to trigger WD	*/

static int device_open = 0;	/* only one application may open device wdt	*/


int wdt_mpc8xx_init (void);
void wdt_mpc8xx_handler (unsigned long);
static int wdt_mpc8xx_open (struct inode *, struct file *);
static int wdt_mpc8xx_release (struct inode *, struct file *);
static ssize_t wdt_mpc8xx_read (struct file *, char *, size_t, loff_t *);
static ssize_t wdt_mpc8xx_write (struct file *, const char *, size_t, loff_t *);
static int wdt_mpc8xx_ioctl (struct inode *, struct file *,
			     unsigned int, unsigned long);

#ifndef KERNEL_VERSION
# define KERNEL_VERSION(a,b,c)	(((a) << 16) + ((b) << 8) + (c))
#endif

#if (!defined(LINUX_VERSION_CODE) || LINUX_VERSION_CODE < KERNEL_VERSION(2,2,0))
#error "You need to use at least Linux kernel version 2.2.0"
#endif

/*
 * Deal with CONFIG_MODVERSIONS
 */
#if CONFIG_MODVERSIONS==1
# include <linux/modversions.h>
#endif

#if LINUX_VERSION_CODE < KERNEL_VERSION(2,3,0)
struct file_operations wdt_mpc8xx_ops = {
	NULL,			/* seek    */
	wdt_mpc8xx_read,
	wdt_mpc8xx_write,
	NULL,			/* readdir */
	NULL,			/* select  */
	wdt_mpc8xx_ioctl,	/* ioctl   */
	NULL,			/* mmap    */
	wdt_mpc8xx_open,
	NULL,			/* flush   */
	wdt_mpc8xx_release
};

#else	/* Linux kernel >= 2.3.x */

struct file_operations wdt_mpc8xx_ops = {
	owner:			THIS_MODULE,
	open:			wdt_mpc8xx_open,
	release:		wdt_mpc8xx_release,
	read:			wdt_mpc8xx_read,
	write:			wdt_mpc8xx_write,
	ioctl:			wdt_mpc8xx_ioctl,
};

#endif

static struct miscdevice wdt_miscdev =	/* driver is a misc device */
{
	WATCHDOG_MINOR,
	"watchdog",
	&wdt_mpc8xx_ops
};


int __init wdt_mpc8xx_init (void)
{
	int rc;

	/* get pointer for SYPCR and SWSR */
	if (!immr) {
		immr = (immap_t *) IMAP_ADDR;
	}

	debugk ("%s: SYPCR=0x%x\n", __FUNCTION__, immr->im_siu_conf.sc_sypcr);

#ifndef	CONFIG_LWMON	/* LWMON uses external MAX708TESA watchdog */
	if ((immr->im_siu_conf.sc_sypcr & 0x00000004) == 0) {
		printk ("WDT_8xx: SWT not enabled by firmware, SYPCR=0x%x\n",
			immr->im_siu_conf.sc_sypcr);
		return (-ENXIO);
	}
#endif

	/* register misc device */
	if ((rc = misc_register (&wdt_miscdev)) < 0) {
		printk ("%s: failed with %d\n", __FUNCTION__, rc);
		return rc;
	}

	debugk ("WDT_8xx registered: major=%d minor=%d\n",
		MISC_MAJOR, WATCHDOG_MINOR);

#ifdef CONFIG_LWMON	/* LWMON uses external MAX708TESA watchdog */

	immr->im_ioport.iop_padat ^= 0x0100;

#else			/* use MPC8xx internal SWT */
	/* set SYPCR[SWRI]...wdt-timeout causes reset */
	immr->im_siu_conf.sc_sypcr |= (
				0x00000002	/* SWRI - SWT causes HRESET  */
				|
				0x00000001	/* SWP  - activate prescaler */
				|
				0xFFFF0000	/* SWTC - max. timer count   */
				);

	debugk ("%s: SYPCR=0x%x\n", __FUNCTION__, immr->im_siu_conf.sc_sypcr);

	/* trigger now */
	immr->im_siu_conf.sc_swsr = 0x556C;
	immr->im_siu_conf.sc_swsr = 0xAA39;
#endif

	debugk ("%s: watchdog timer initialized - timer_count = %d  period = %d\n",
		__FUNCTION__, timer_count / HZ, HZ / timer_period);

	init_timer (&wd_timer);			/* initialize timer-structure... */
	wd_timer.function = wdt_mpc8xx_handler;
	wd_timer.data = (unsigned long) &wdt_mpc8xx_dummy;
	wd_timer.expires = jiffies + timer_period;

	add_timer (&wd_timer);			/* ...and activate timer */

	debugk ("%s: timer activated\n", __FUNCTION__);

	printk ("WDT_8xx: Software Watchdog Timer version " WDT_VERSION);

	if (enabled) {
		printk (", timeout %d sec.\n", timer_count / HZ);
	} else {
		printk (" (disabled)\n");
	}

	return 0;
}


int __init wdt_mpc8xx_setup (char *options)
{
	while (options && *options) {
		if (strncmp (options, "off", 3) == 0) {
			options += 3;
			enabled = 0;
			if (*options != ',')
				return 0;
			options++;
		}

		if (strncmp (options, "timeout:", 8) == 0) {
			options += 8;
			if (!*options)
				return 0;

			timer_count = HZ * simple_strtoul (options, &options, 0);
			if (*options != ',')
				return 0;
			options++;
		}

		if (strncmp (options, "period:", 7) == 0) {
			options += 7;
			if (!*options)
				return 0;
			
			timer_period = simple_strtoul (options, &options, 0);
			if (*options != ',')
				return 0;
			options++;
		}
	}

	return 0;
}

__setup ("wdt_8xx=", wdt_mpc8xx_setup);


static int wdt_mpc8xx_open (struct inode *inode, struct file *filp)
{
	debugk ("ENTER %s (%p, %p)\n", __FUNCTION__, inode, filp);

	if (!enabled) {			/* user interface disabled */
		return -ENXIO;
	}
	if (device_open) {		/* exclusive open only */
		debugk ("%s: WDT is busy\n", __FUNCTION__);
		return -EBUSY;
	}
	device_open++;			/* increment usage counter */
	MOD_INC_USE_COUNT;
	debugk ("%s: WDT is open\n", __FUNCTION__);
	return 0;
}


static int wdt_mpc8xx_release (struct inode *inode, struct file *filp)
{
	debugk ("ENTER %s (%p, %p)\n", __FUNCTION__, inode, filp);

	device_open--;			/* decrement usage counter */
	MOD_DEC_USE_COUNT;
	debugk ("%s: WDT is closed\n", __FUNCTION__);

	return 0;
}


static ssize_t wdt_mpc8xx_read (struct file *filp, char *buffer,
				size_t length,   loff_t *offset)
{
	unsigned int rest_count = timer_count / HZ;

	debugk ("ENTER %s (%p, %p, %d, %p)\n",
		__FUNCTION__, filp, buffer, length,  offset);

	if (length < sizeof(rest_count)) {
		debugk ("wdt_mpc8xx_release/kernel: invalid argument\n");

		return -EINVAL;
	}
	put_user (rest_count, (int *) buffer);	/* copy value into userspace */

	debugk ("%s: rest_count=%i\n", __FUNCTION__, rest_count);

	return (sizeof(rest_count));	/* read always exactly 4 bytes */
}


static ssize_t wdt_mpc8xx_write (struct file *filp, const char *buffer,
				 size_t length, loff_t *offset)
{
	int error;
	unsigned int new_count;

	debugk ("ENTER %s (%p, %p, %d, %p)\n",
		__FUNCTION__, filp, buffer, length,  offset);

	if (length != sizeof(new_count)) {
		debugk ("%s: invalid length (%d instead of %d)\n",
			__FUNCTION__, length, sizeof(new_count));

		return -EINVAL;
	}

	/* copy count value into kernel space */
	if ((error = get_user (new_count, (int *) buffer)) != 0) {
		debugk ("%s: get_user failed: rc=%d\n",
			__FUNCTION__, error);

		return (error);
	}

	timer_count = new_count * HZ;

	return (sizeof(new_count));
}

static int wdt_mpc8xx_ioctl (struct inode *node, struct file *filp,
			     unsigned int cmd, unsigned long arg)
{
	switch (cmd) {
	case WDT_MPC8XX_OPEN_ONLY:
			timeout_open_only = 1;
			break;
	case WDT_MPC8XX_ALWAYS:
			timeout_open_only = 0;
			break;
	default:
			return -EINVAL;
	}
	return 0;
}


#ifdef MODULE

int init_module (void)
{
	debugk ("%s: initialize WDT8xx\n", __FUNCTION__);

	return wdt_mpc8xx_init ();
}


void cleanup_module (void)
{
	debugk ("%s: cleanup WDT8xx\n", __FUNCTION__);

	del_timer (&wd_timer);

	misc_deregister (&wdt_miscdev);
}

#endif	/* MODULE */


void wdt_mpc8xx_handler (unsigned long ptr)
{

	debugk ("%s: timer_count=%d jiffies\n", __FUNCTION__, timer_count);

	if ((timer_count == 0) && enabled) {	/* try "graceful" shutdown */
		extern void m8xx_restart(char *);

		printk ("WDT_8xx: Reset system...\n");
		m8xx_restart (NULL);
	}

	if ((timer_count > 0) || (!enabled)) {

		/* execute WDT service sequence */
# if defined(CONFIG_LWMON)
		/*
		 * The LWMON board uses a MAX706TESA Watchdog
		 * with the trigger pin connected to port PA.7
		 *
		 * The port has already been set up in the firmware,
		 * so we just have to toggle it.
		 */
		immr->im_ioport.iop_padat ^= 0x0100;
# else
		/*
		 * All other boards use the MPC8xx Internal Watchdog
		 */
		immr->im_siu_conf.sc_swsr = 0x556C;
		immr->im_siu_conf.sc_swsr = 0xAA39;
		debugk ("%s: WDT serviced\n", __FUNCTION__);
# endif /* CONFIG_LWMON */

		wd_timer.expires = jiffies + timer_period;
		add_timer (&wd_timer);	/* ...re-activate timer */

		/*
		 * don't timeout if disabled
		 */
		if (!enabled)
			return;

		/*
		 * don't timeout if device isn't open
		 */
		if ((timeout_open_only) && (! device_open))
			return;

		/* decrement variable for timer-control */
		if (timer_count > timer_period)
			timer_count -= timer_period;
		else
			timer_count = 0;

		if (timer_count == 0) {
			printk ("WDT_8xx: watchdog about to expire\n");
		}
	}

	return;
}
