/*
 * (C) Copyright 2000
 * Jrg Haider, SIEMENS AG
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License as
 * published by the Free Software Foundation; either version 2 of
 * the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston,
 * MA 02111-1307 USA
 *
 * (C) 2001 Wolfgang Denk, wd@denx.de -- Cleanup, Modifications for 2.4 kernels
 * (C) 2001 Wolfgang Denk, wd@denx.de -- Adaption for MAX706TESA Watchdog (LWMON)
 * (C) 2001 Steven Hein,  ssh@sgi.com -- Added timeout configuration option
 * (C) 2001 Robert Enzinger, Robert.Enzinger.extern@icn.siemens.de,
 *		-- added ioctl() for dynamic configuration
 */

#include <linux/config.h>
#include <linux/module.h>
#include <linux/version.h>
#include <linux/types.h>
#include <linux/sched.h>
#include <linux/kernel.h>
#include <linux/errno.h>
#include <linux/miscdevice.h>		/* driver is a misc device	*/
#include <linux/init.h>			/* for __initfunc		*/
#include <asm/wdt_pcippc2.h>
#include <asm/uaccess.h>		/* for put_user			*/
#include <asm/pcippc2.h>
#include <asm/io.h>
#include <linux/fs.h>			/* for character devices	*/

#undef DEBUG
#define CONFIG_WDT_PCIPPC_TIMEOUT_OPEN_ONLY

#ifdef DEBUG
# define debugk(fmt,args...) printk(fmt ,##args)
#else
# define debugk(fmt,args...)
#endif

#define	WDT_VERSION	"0.1"	/* version of this device driver		*/

#ifdef	CONFIG_WDT_PCIPPC_TIMEOUT
#define TIMEOUT_VALUE CONFIG_WDT_PCIPPC_TIMEOUT	/* configurable timeout		*/
#else
#define TIMEOUT_VALUE 300	/* reset after five minutes = 300 seconds	*/
#endif

#ifdef CONFIG_WDT_PCIPPC_TIMEOUT_OPEN_ONLY
static int timeout_open_only = 1;
#else
static int timeout_open_only = 0;
#endif

struct timer_list wd_timer;	/* structure for timer administration		*/
static unsigned long wdt_pcippc2_dummy;

/*
 * Watchdog active? When disabled, it will get re-triggered
 * automatically without timeout, so it appears to be switched off
 * although actually it is still running.
 */
static int enabled = 1;

/*
 * The shortest watchdog period of all boards is (so far) approx. 1 sec,
 * thus re-trigger watchdog by default every 500 ms = HZ / 2.
 * The default can be overwritten using a boot argument.
 *
 * The external interface is in seconds, but internal all calculation
 * is done in jiffies.
 */
#define	TRIGGER_PERIOD	( HZ / 2 )
static unsigned long timer_count = TIMEOUT_VALUE * HZ;	/* remaining time	*/
static unsigned long timer_period= TRIGGER_PERIOD;	/* period to trigger WD	*/

static int device_open = 0;	/* only one application may open device wdt	*/


int wdt_pcippc2_init (void);
void wdt_pcippc2_handler (unsigned long);
static int wdt_pcippc2_open (struct inode *, struct file *);
static int wdt_pcippc2_release (struct inode *, struct file *);
static ssize_t wdt_pcippc2_read (struct file *, char *, size_t, loff_t *);
static ssize_t wdt_pcippc2_write (struct file *, const char *, size_t, loff_t *);
static int wdt_pcippc2_ioctl (struct inode *, struct file *,
			     unsigned int, unsigned long);

#ifndef KERNEL_VERSION
# define KERNEL_VERSION(a,b,c)	(((a) << 16) + ((b) << 8) + (c))
#endif

#if (!defined(LINUX_VERSION_CODE) || LINUX_VERSION_CODE < KERNEL_VERSION(2,2,0))
#error "You need to use at least Linux kernel version 2.2.0"
#endif

#if LINUX_VERSION_CODE < KERNEL_VERSION(2,3,0)
struct file_operations wdt_pcippc2_ops = {
	NULL,			/* seek    */
	wdt_pcippc2_read,
	wdt_pcippc2_write,
	NULL,			/* readdir */
	NULL,			/* select  */
	wdt_pcippc2_ioctl,	/* ioctl   */
	NULL,			/* mmap    */
	wdt_pcippc2_open,
	NULL,			/* flush   */
	wdt_pcippc2_release
};

#else	/* Linux kernel >= 2.3.x */

struct file_operations wdt_pcippc2_ops = {
	owner:			THIS_MODULE,
	open:			wdt_pcippc2_open,
	release:		wdt_pcippc2_release,
	read:			wdt_pcippc2_read,
	write:			wdt_pcippc2_write,
	ioctl:			wdt_pcippc2_ioctl,
};

#endif

static struct miscdevice wdt_miscdev =	/* driver is a misc device */
{
	WATCHDOG_MINOR,
	"watchdog",
	&wdt_pcippc2_ops
};


int __init wdt_pcippc2_init (void)
{
	int rc, ctl;

	ctl = inb(FPGA(WDT,CTRL));

	debugk ("%s: =0x%x\n", __FUNCTION__, ctl);

	/* register misc device */
	if ((rc = misc_register (&wdt_miscdev)) < 0) {
		printk ("%s: failed with %d\n", __FUNCTION__, rc);
		return rc;
	}

	debugk ("WDT_pcippc registered: major=%d minor=%d\n",
		MISC_MAJOR, WATCHDOG_MINOR);

	outw(0xffff, FPGA(WDT,PROG));	/* max. timer count   */
	outb(0x56, FPGA(WDT,REFRESH));	/* refresh timer value */
	outb(1, FPGA(WDT,CTRL));

	debugk ("%s: watchdog timer initialized - timer_count = %ld  period = %ld\n",
		__FUNCTION__, timer_count / HZ, HZ / timer_period);

	init_timer (&wd_timer);			/* initialize timer-structure... */
	wd_timer.function = wdt_pcippc2_handler;
	wd_timer.data = (unsigned long) &wdt_pcippc2_dummy;
	wd_timer.expires = jiffies + timer_period;

	add_timer (&wd_timer);			/* ...and activate timer */

	debugk ("%s: timer activated\n", __FUNCTION__);

	printk ("WDT_pcippc: Software Watchdog Timer version " WDT_VERSION);

	if (enabled) {
		printk (", timeout %ld sec.\n", timer_count / HZ);
	} else {
		printk (" (disabled)\n");
	}

	return 0;
}


int __init wdt_pcippc2_setup (char *options)
{
	while (options && *options) {
		if (strncmp (options, "off", 3) == 0) {
			options += 3;
			enabled = 0;
			if (*options != ',')
				return 0;
			options++;
		}

		if (strncmp (options, "timeout:", 8) == 0) {
			options += 8;
			if (!*options)
				return 0;

			timer_count = HZ * simple_strtoul (options, &options, 0);
			if (*options != ',')
				return 0;
			options++;
		}

		if (strncmp (options, "period:", 7) == 0) {
			options += 7;
			if (!*options)
				return 0;
			
			timer_period = simple_strtoul (options, &options, 0);
			if (*options != ',')
				return 0;
			options++;
		}
	}

	return 0;
}

__setup ("wdt_pcippc2=", wdt_pcippc2_setup);


static int wdt_pcippc2_open (struct inode *inode, struct file *filp)
{
	debugk ("ENTER %s (%p, %p)\n", __FUNCTION__, inode, filp);

	if (!enabled) {			/* user interface disabled */
		return -ENXIO;
	}
	if (device_open) {		/* exclusive open only */
		debugk ("%s: WDT is busy\n", __FUNCTION__);
		return -EBUSY;
	}
	device_open++;			/* increment usage counter */
	MOD_INC_USE_COUNT;
	debugk ("%s: WDT is open\n", __FUNCTION__);
	return 0;
}


static int wdt_pcippc2_release (struct inode *inode, struct file *filp)
{
	debugk ("ENTER %s (%p, %p)\n", __FUNCTION__, inode, filp);

	device_open--;			/* decrement usage counter */
	MOD_DEC_USE_COUNT;
	debugk ("%s: WDT is closed\n", __FUNCTION__);

	return 0;
}


static ssize_t wdt_pcippc2_read (struct file *filp, char *buffer,
				size_t length,   loff_t *offset)
{
	unsigned int rest_count = timer_count / HZ;

	debugk ("ENTER %s (%p, %p, %d, %p)\n",
		__FUNCTION__, filp, buffer, length,  offset);

	if (length < sizeof(rest_count)) {
		debugk ("wdt_pcippc2_release/kernel: invalid argument\n");

		return -EINVAL;
	}
	put_user (rest_count, (int *) buffer);	/* copy value into userspace */

	debugk ("%s: rest_count=%i\n", __FUNCTION__, rest_count);

	return (sizeof(rest_count));	/* read always exactly 4 bytes */
}


static ssize_t wdt_pcippc2_write (struct file *filp, const char *buffer,
				 size_t length, loff_t *offset)
{
	int error;
	unsigned int new_count;

	debugk ("ENTER %s (%p, %p, %d, %p)\n",
		__FUNCTION__, filp, buffer, length,  offset);

	if (length != sizeof(new_count)) {
		debugk ("%s: invalid length (%d instead of %d)\n",
			__FUNCTION__, length, sizeof(new_count));

		return -EINVAL;
	}

	/* copy count value into kernel space */
	if ((error = get_user (new_count, (int *) buffer)) != 0) {
		debugk ("%s: get_user failed: rc=%d\n",
			__FUNCTION__, error);

		return (error);
	}

	timer_count = new_count * HZ;

	return (sizeof(new_count));
}

static int wdt_pcippc2_ioctl (struct inode *node, struct file *filp,
			     unsigned int cmd, unsigned long arg)
{
	switch (cmd) {
	case WDT_PCIPPC_OPEN_ONLY:
			timeout_open_only = 1;
			break;
	case WDT_PCIPPC_ALWAYS:
			timeout_open_only = 0;
			break;
	default:
			return -EINVAL;
	}
	return 0;
}


#ifdef MODULE

int init_module (void)
{
	debugk ("%s: initialize WDT_pcippc\n", __FUNCTION__);

	return wdt_pcippc2_init ();
}


void cleanup_module (void)
{
	debugk ("%s: cleanup WDT_pcippc\n", __FUNCTION__);

	del_timer (&wd_timer);

	misc_deregister (&wdt_miscdev);
}

#endif	/* MODULE */


void wdt_pcippc2_handler (unsigned long ptr)
{

	debugk ("%s: timer_count=%ld jiffies\n", __FUNCTION__, timer_count);

	if ((timer_count == 0) && enabled) {	/* try "graceful" shutdown */
		extern void pcippc2_restart(char *);

		printk ("WDT_pcippc: Reset system...\n");
		pcippc2_restart (NULL);
	}

	if ((timer_count > 0) || (!enabled)) {

		/* execute WDT service sequence */
		outb(0x56, FPGA(WDT, REFRESH));

		wd_timer.expires = jiffies + timer_period;
		add_timer (&wd_timer);	/* ...re-activate timer */

		/*
		 * don't timeout if disabled
		 */
		if (!enabled)
			return;

		/*
		 * don't timeout if device isn't open
		 */
		if ((timeout_open_only) && (! device_open))
			return;

		/* decrement variable for timer-control */
		if (timer_count > timer_period)
			timer_count -= timer_period;
		else
			timer_count = 0;

		if (timer_count == 0) {
			printk ("WDT_pcippc: watchdog about to expire\n");
		}
	}

	return;
}
