/***********************************************************************
 *
 * (C) Copyright 2001
 * DENX Software Engineering
 * Wolfgang Denk, wd@denx.de
 * All rights reserved.
 *
 * Simple IC driver for LWMON
 * Modeled after
 *	i2c-algo-8xx.c i2c driver algorithms for MPC8XX CPM
 *	Copyright (c) 1999 Dan Malek (dmalek@jlc.net).
 *
 *
 * We do not provide read/write system calls for user applications;
 * all access will be done using ioctl() calls.
 *
 * This is necessary for easy synchronization of multiple parallel
 * accesses to differen IC addresses.
 *
 ***********************************************************************/

#include <linux/kernel.h>
#include <linux/module.h>
#include <linux/delay.h>
#include <linux/malloc.h>
#include <linux/init.h>
#include <linux/fs.h>
#include <linux/ioport.h>
#include <linux/errno.h>
#include <linux/sched.h>
#include <linux/version.h>
#include <linux/devfs_fs_kernel.h>
#include <asm/semaphore.h>
#include <asm/uaccess.h>

#include <asm/mpc8xx.h>
#include "../../arch/ppc/8xx_io/commproc.h"

#include "i2c-simple.h"

/*
 * Debugging Options
 *
 * Warning - the use of I2C_SCAN can cause problems when accessing the devices
 */
/* #define I2C_SCAN	1	// */
/* #define I2C_DEBUG	1	//	*/

/*
 * Deal with CONFIG_MODVERSIONS
 */
#if CONFIG_MODVERSIONS==1
/* # define MODVERSIONS */
# include <linux/modversions.h>
#endif

#ifdef MODULE
MODULE_AUTHOR("Wolfgang Denk, wd@denx.de");
MODULE_DESCRIPTION("Simple I2C Driver for LWMON Project");
#endif	/* MODULE */

/*
 * The name for our device, as it will appear in /proc/devices
 */
#define DEVICE_NAME "i2c-simple"


#ifndef I2C_MAJOR
#define I2C_MAJOR	89		/* I2C Major Device Number	*/
#endif

/*
 * Maximum number of times to poll for ACK after EEPROM write.
 */
#define MAX_ACKNOWLEDGE_POLLS 10

typedef struct si2c_data {
	uint	dp_addr;
	int	reloc;
	volatile i2c8xx_t	*i2c;
	volatile iic_t		*iip;
} si2c_data_t;

static si2c_data_t si2c_data;

static wait_queue_head_t iic_wait;
static ushort r_tbase, r_rbase;

/*
 * allow many processes to open the device,
 * but serialize system call [i. e. ioctl()] access to CPM
 */
static struct semaphore i2c_sem;

/**********************************************************************/

#ifdef MODULE
int  init_module(void);
void cleanup_module(void);
#endif

int	__init	si2c_init (void);
static int	si2c_open (struct inode *inode, struct file *file);
static int	si2c_release (struct inode *inode, struct file *file);
static ssize_t	si2c_ioctl (struct inode *inode, struct file *file,
			    unsigned int cmd, unsigned long arg);
static void	si2c_interrupt (void *dev_id);
static int	si2c_shutdown (void);
static void	si2c_reset_params (volatile iic_t *iip);
static void	si2c_force_close (void);
#ifdef I2C_SCAN
static int	si2c_tryaddress (int addr);
#endif	/* I2C_SCAN */

/* imported from arch/ppc/kernel/misc.S */
extern void invalidate_dcache_range(ulong start, ulong stop);

/**********************************************************************/

static struct file_operations si2c_ops =
{
	owner:		THIS_MODULE,
	open:		si2c_open,
	release:	si2c_release,
	ioctl:		si2c_ioctl,
};

/**********************************************************************/


#ifdef MODULE

int init_module (void)
{
	return si2c_init ();
}

void cleanup_module (void)
{
	int rc;

#ifdef I2C_DEBUG
	printk ("si2c cleanup_module()\n");
#endif

	/* Uninstall interrupt handler */
	cpm_free_handler (CPMVEC_I2C);

	/*
	 * Unregister the device
	 */
	rc = unregister_chrdev (I2C_MAJOR, DEVICE_NAME);

	/*
	 * If there's an error, report it
	 */
	if (rc < 0) {
		printk ("unregister_chrdev (%s): error %d\n",
			DEVICE_NAME, rc);
	}
}

#endif	/* MODULE */

/**********************************************************************/

int __init si2c_init (void)
{
	volatile cpm8xx_t *cp = cpmp;	/* exported by commproc.c */
	volatile immap_t  *immap = (immap_t *)IMAP_ADDR;
	volatile iic_t	  *iip;
	volatile i2c8xx_t *i2c;
	si2c_data_t *data = &si2c_data;
	int n;

#ifdef I2C_DEBUG
	printk ("si2c_init()\n");
#endif

	if ((n = register_chrdev (I2C_MAJOR, DEVICE_NAME, &si2c_ops)) < 0) {
		printk (DEVICE_NAME ": init_module failed with %d\n", n);
		return (n);
	}

	printk (KERN_INFO DEVICE_NAME ": $Revision: 1.1.1.1 $ wd@denx.de\n");

	i2c = data->i2c = (i2c8xx_t *)&(immap->im_i2c);
	iip = data->iip = (iic_t *)&cp->cp_dparam[PROFF_IIC];

#ifdef USE_IIC_PATCH
	/* Check for and use a microcode relocation patch.
	*/
	if ((data->reloc = data->iip->iic_rpbase))
		iip = data->iip = (iic_t *)&cp->cp_dpmem[data->iip->iic_rpbase];
#endif

	/* Initialize Port B I2C pins.
	*/
	cp->cp_pbpar |= 0x00000030;
	cp->cp_pbdir |= 0x00000030;
	cp->cp_pbodr |= 0x00000030;

	/* Allocate space for two transmit and two receive buffer
	 * descriptors in the DP ram.
	 */
	data->dp_addr = m8xx_cpm_dpalloc (sizeof(cbd_t) * 4);

	/* Initialize the parameter ram.
	 * We need to make sure many things are initialized to zero,
	 * especially in the case of a microcode patch.
	 */
	iip->iic_rstate = 0;
	iip->iic_rdp = 0;
	iip->iic_rbptr = 0;
	iip->iic_rbc = 0;
	iip->iic_rxtmp = 0;
	iip->iic_tstate = 0;
	iip->iic_tdp = 0;
	iip->iic_tbptr = 0;
	iip->iic_tbc = 0;
	iip->iic_txtmp = 0;

	/* Set up the I2C parameters in the parameter ram.
	 */
	iip->iic_tbase = r_tbase = si2c_data.dp_addr;
	iip->iic_rbase = r_rbase = si2c_data.dp_addr + sizeof(cbd_t)*2;

	iip->iic_tfcr = SMC_EB;
	iip->iic_rfcr = SMC_EB;

	/* Set maximum receive size.
	 */
	iip->iic_mrblr = I2C_MAX_LEN;

	/* Initialize Tx/Rx parameters.
	 */
	if (si2c_data.reloc == 0) {
		cp->cp_cpcr =
			mk_cr_cmd (CPM_CR_CH_I2C, CPM_CR_INIT_TRX) | CPM_CR_FLG;
		while (cp->cp_cpcr & CPM_CR_FLG);
	}

	/* Select an arbitrary address.	 Just make sure it is unique.
	 */
	i2c->i2c_i2add = 0x34;

	/* Make clock run maximum slow.
	 */
	i2c->i2c_i2brg = 7;

	/* Disable interrupts.
	 */
	i2c->i2c_i2cmr = 0;
	i2c->i2c_i2cer = 0xff;

	init_waitqueue_head (&iic_wait);

	init_MUTEX(&i2c_sem);

	/* Install interrupt handler.
	 */
#ifdef I2C_DEBUG
	printk ("%s[%d] Install ISR for IRQ %d\n",
		__FUNCTION__, __LINE__, CPMVEC_I2C);
#endif
	/* install interrupt handler
	 */
	cpm_install_handler (CPMVEC_I2C, si2c_interrupt, (void *)i2c);

#ifdef I2C_SCAN
	/*
	 * Scan I2C bus
	 *
	 * Warning - this can cause problems for the following access(es)
	 * to the I2C devices.
	 */
	printk (KERN_INFO "Scanning I2C bus:");
	for (n = 0; n < 128; n++) {
		if (si2c_tryaddress (n)) {
			printk (" %02X", i);
		}
	}
	printk ("\n");
#endif	/* I2C_SCAN */

	return (0);
}

/**********************************************************************/

static int si2c_open (struct inode *inode, struct file *file)
{
#ifdef I2C_DEBUG
	printk (__FUNCTION__ "() called\n");
#endif

	/*
	 * Get major / minor numbers when needed
	 */
#ifdef I2C_DEBUG
	printk (DEVICE_NAME ": Device: major=%d minor=%d\n",
		inode->i_rdev >> 8, inode->i_rdev & 0xFF);
#endif

	MOD_INC_USE_COUNT;

	return (0);
}

/**********************************************************************/

static int si2c_release (struct inode *inode, struct file *file)
{
#ifdef I2C_DEBUG
	printk (__FUNCTION__ "() called\n");
#endif

	si2c_shutdown ();

	MOD_DEC_USE_COUNT;

	return (0);
}

/**********************************************************************/

/*
 * Read from I2C:
 *
 * The parameter structure passes the following information:
 *
 * alen:	Number of address bytes in "addr"
 * addr:	Array of address bytes; for instance, to read from
 *		EEPROM with 8 bit page addresses you have to write 2
 *		address bytes: the block number, and the block
 *		offset.
 * dlen:	number of data bytes to read
 *		Warning: it is expected that the device is really
 *		capable of reading `dlen' sequential bytes; no
 *		checking is done here
 * data:	where to store the read data in
 */

EXPORT_SYMBOL(si2c_read);
ssize_t si2c_read (si2c_request_t *request)
{
	volatile iic_t    *iip = si2c_data.iip;
	volatile i2c8xx_t *i2c = si2c_data.i2c;
	volatile cpm8xx_t *cp  = cpmp;
	volatile cbd_t	*tbdf, *rbdf;
	unsigned long flags;
	u_char	temp[I2C_MAX_LEN];
	u_char *tb;
	int i;

#ifdef I2C_DEBUG
	printk ("si2c_read: alen: %d addr:", request->alen);
	for (i=0; i<request->alen; ++i) {
		printk (" %02X", request->addr[i]);
	}
	printk (" dlen: %d\n", request->dlen);
#endif

	if (request->dlen >= I2C_MAX_LEN)
		return (-EINVAL);

	/* check for and use a microcode relocation patch */
	if (si2c_data.reloc) {
		si2c_reset_params (iip);
	}

	tbdf = (cbd_t *)&cp->cp_dpmem[iip->iic_tbase];
	rbdf = (cbd_t *)&cp->cp_dpmem[iip->iic_rbase];

        /*
         * To read, we need an empty buffer of the proper length. All
         * that is used is the first few bytes for address, the
         * remainder is just used for timing (and doesn't really have
         * to exist).
	 */
	if (si2c_data.reloc) {
		si2c_reset_params (iip);
	}
	tb = temp;
	tb = (u_char *)(((uint)tb + 15) & ~15);

	/* Shift address, and insert read flag bit
	 */
	tb[0] = (request->addr[0] << 1) | 0x01;
	for (i=1; i<request->alen; ++i) {
		tb[i] = request->addr[i];
	}

	flush_dcache_range ((ulong)tb, (ulong)(tb + request->alen));

	tbdf->cbd_bufaddr = __pa(tb);
	tbdf->cbd_datlen  = request->dlen + request->alen;
	tbdf->cbd_sc	  = BD_SC_READY | BD_SC_INTRPT | BD_SC_LAST |
			    BD_SC_WRAP  | BD_IIC_START;

	rbdf->cbd_datlen = 0;
	rbdf->cbd_bufaddr = __pa(request->data);
	rbdf->cbd_sc = BD_SC_EMPTY | BD_SC_WRAP;

#ifdef CONFIG_LWMON
	if (request->dlen > 0 && request->dlen < I2C_MAX_LEN)
		iip->iic_mrblr = request->dlen; /* prevent excessive read */
#endif

	invalidate_dcache_range ((ulong) request->data,
				 (ulong)(request->data + request->dlen));

	/* Chip bug, set enable here */
	save_flags (flags); cli ();
	i2c->i2c_i2cmr = 0x13;	/* Enable some interupts */
	i2c->i2c_i2cer = 0xFF;
	i2c->i2c_i2mod = 1;	/* Enable */
	i2c->i2c_i2com = 0x81;	/* Start master */

	/* Wait for I2C transfer */
	interruptible_sleep_on (&iic_wait);

	restore_flags (flags);

	if (signal_pending (current)) {
		return (-EIO);
	}

#ifdef I2C_DEBUG
	printk ("I2C read: tx sc %04x, rx sc %04x\n",
	       tbdf->cbd_sc, rbdf->cbd_sc);
#endif

	if (tbdf->cbd_sc & BD_SC_NAK) {
#ifndef CONFIG_ICU862	/* must use polling on X40430/40431 chips */
		printk ("I2C read; no ack\n");
#endif
		return (0);
	}

	if (rbdf->cbd_sc & BD_SC_EMPTY) {
#ifndef CONFIG_ICU862	/* must use polling on X40430/40431 chips */
		printk ("I2C read; complete but rbuf empty\n");
#endif

		si2c_force_close ();

#ifndef CONFIG_ICU862	/* must use polling on X40430/40431 chips */
		printk ("I2C read: tx sc %04x, rx sc %04x\n",
		       tbdf->cbd_sc, rbdf->cbd_sc);
#endif
	}

#ifdef I2C_DEBUG
	printk ("I2C read: got %d bytes\n", rbdf->cbd_datlen);
#endif

	if (rbdf->cbd_datlen < request->dlen) {
		printk ("I2C read: short, wanted %d got %d\n",
		       request->dlen, rbdf->cbd_datlen);
		request->dlen = rbdf->cbd_datlen;
	}

#ifdef I2C_DEBUG
	printk ("I2C read: dlen %d data:", request->dlen);
	for (i=0; i<request->dlen; ++i) {
		printk (" %02X", request->data[i]);
	}
	printk ("\n");
#endif

	return (request->dlen);
}

/**********************************************************************/

/*
 * Write to I2C...
 *
 * like reading...
 */
EXPORT_SYMBOL(si2c_write);
ssize_t si2c_write (si2c_request_t *request)
{
	volatile iic_t    *iip = si2c_data.iip;
	volatile i2c8xx_t *i2c = si2c_data.i2c;
	volatile cpm8xx_t *cp  = cpmp;
	volatile cbd_t	*tbdf;
	unsigned long flags;
	u_char	temp[I2C_MAX_LEN];
	u_char *tb;
	int i;

#ifdef I2C_DEBUG
	printk ("si2c_write: alen: %d addr:", request->alen);
	for (i=0; i<request->alen; ++i) {
		printk (" %02X", request->addr[i]);
	}
	printk (" dlen: %d data:", request->dlen);
	for (i=0; i<request->dlen; ++i) {
		printk (" %02X", request->data[i]);
	}
	printk ("\n");
#endif

	/* check for and use a microcode relocation patch */
	if (si2c_data.reloc) {
		si2c_reset_params (iip);
	}
	tb = temp;
	tb = (u_char *)(((uint)tb + 15) & ~15);

	/* Shift address, and insert write flag bit
	 */
	tb[0] = (request->addr[0] << 1) | 0x00;
	for (i=1; i<request->alen; ++i) {
		tb[i] = request->addr[i];
	}

	flush_dcache_range ((ulong)tb,
			    (ulong)(tb + request->alen));
	flush_dcache_range ((ulong)request->data,
			    (ulong)(request->data + request->dlen));

	/* set up 2 descriptors */
	tbdf = (cbd_t *)&cp->cp_dpmem[iip->iic_tbase];

	tbdf[0].cbd_bufaddr = __pa(tb);
	tbdf[0].cbd_datlen = 1;
	tbdf[0].cbd_sc = BD_SC_READY | BD_IIC_START;

	tbdf[1].cbd_bufaddr = __pa(request->data);
	tbdf[1].cbd_datlen = request->dlen;
	tbdf[1].cbd_sc = BD_SC_READY | BD_SC_INTRPT | BD_SC_LAST | BD_SC_WRAP;

	/* Chip bug, set enable here */
	save_flags (flags); cli ();
	i2c->i2c_i2cmr = 0x13;	/* Enable some interupts */
	i2c->i2c_i2cer = 0xFF;
	i2c->i2c_i2mod = 1;	/* Enable */
	i2c->i2c_i2com = 0x81;	/* Start master */

	/* Wait for I2C transfer */
	interruptible_sleep_on (&iic_wait);

	restore_flags (flags);

	if (signal_pending (current)) {
		return (-EIO);
	}

#ifdef I2C_DEBUG
	printk ("I2C write: tx0 sc %04x, tx1 sc %04x\n",
	       tbdf[0].cbd_sc, tbdf[1].cbd_sc);
#endif

	if (tbdf->cbd_sc & BD_SC_NAK) {
		printk ("I2C write: no ack\n");
		return 0;
	}

	if (tbdf->cbd_sc & BD_SC_READY) {
		printk ("I2C write: complete but tbuf ready\n");
		return 0;
	}

	return (request->dlen);
}

/**********************************************************************/

static ssize_t si2c_ioctl (struct inode *inode, struct file *file,
			   unsigned int cmd, unsigned long arg)
{
	si2c_request_t req;
	pic_request_t  pic;
#ifdef CONFIG_ICU862
	eeprom_request_t eeprom;
	unsigned char ctrl_reg;
	int i;
#endif
	ssize_t len;
	int rc = -EINVAL;

	/*
	 * serialize access to CPM
	 */
	if (down_interruptible (&i2c_sem))
		return (-EINTR);

	switch (cmd) {
	default:
		break;

	case I2C_CMD_READ:
	case I2C_CMD_WRITE:
		if (copy_from_user (&req, (si2c_request_t *)arg,
				    sizeof(si2c_request_t))) {
			rc = -EFAULT;
			break;
		}
#ifdef I2C_DEBUG
		printk ("I2C %s [%02x] al=%d dl=%d\n",
			(cmd == I2C_CMD_READ) ? "read" : "write",
			req.addr[0], req.alen, req.dlen );
#endif

		len =  (cmd == I2C_CMD_READ) ? si2c_read(&req) : si2c_write(&req);

		if (len < 0) {
			rc = len;
			break;
		}

		req.dlen = len;

		if (copy_to_user((void *)arg, &req, sizeof(si2c_request_t))) {
			rc = -EFAULT;
			break;
		}

		rc = 0;
		break;

#ifdef CONFIG_LWMON

	case I2C_CMD_PIC_RD:
		if (copy_from_user (&pic, (pic_request_t *)arg,
				    sizeof(pic_request_t))) {
			rc = -EFAULT;
			break;
		}
#ifdef I2C_DEBUG
		printk ("PIC read reg %02X\n", pic.reg);
#endif

		req.addr[0] = I2C_ADDR_PICIO;
		req.alen    = 1;
		req.data[0] = pic.reg;
		req.dlen    = 1;

		if ((len = si2c_write (&req)) < 0) {
			rc = len;
			break;
		}

		req.addr[0] = I2C_ADDR_PICIO;
		req.alen    = 1;
		req.dlen    = 1;

		if ((len = si2c_read (&req)) < 0) {
			rc = len;
			break;
		}

		pic.val = req.data[0];
#ifdef I2C_DEBUG
		printk ("PIC read reg %02X = %02X\n", pic.reg, pic.val);
#endif

		if (copy_to_user((void *)arg, &pic, sizeof(pic_request_t))) {
			rc = -EFAULT;
			break;
		}

		rc = 0;
		break;

	case I2C_CMD_PIC_WR:
		if (copy_from_user (&pic, (pic_request_t *)arg,
				    sizeof(pic_request_t))) {
			rc = -EFAULT;
			break;
		}
#ifdef I2C_DEBUG
		printk ("PIC write reg %02X: val %02X\n", pic.reg, pic.val);
#endif

		req.addr[0] = I2C_ADDR_PICIO;
		req.alen    = 1;
		req.data[0] = pic.reg;
		req.data[1] = pic.val;
		req.dlen    = 2;

		if ((len = si2c_write (&req)) < 0) {
			rc = len;
			break;
		}

		rc = 0;
		break;
#endif	/* CONFIG_LWMON */

#ifdef CONFIG_ICU862

	case I2C_CMD_EEPROM_RD:
		/*
		 * Applies to X40430/40431 chips.
		 */
		if (copy_from_user (&eeprom, (eeprom_request_t *)arg,
				    sizeof(eeprom_request_t))) {
			rc = -EFAULT;
			break;
		}
#ifdef I2C_DEBUG
		printk ("EEPROM read addr %03X\n", eeprom.addr);
#endif

		/* Set current address (internal pointer in the x40430).
		 */
		req.addr[0] = 0x50 | ((eeprom.addr >> 8) & 1);
		req.alen    = 1;
		req.data[0] = (unsigned char) eeprom.addr & 0xff;
		req.dlen    = 1;

		if ((len = si2c_write (&req)) < 0) {
			rc = len;
			break;
		}

		/* A simple read will now extract the byte at that address.
		 */ 
		req.addr[0] = 0x50;
		req.alen    = 1;
		req.dlen    = 1;

		if ((len = si2c_read (&req)) < 0) {
			rc = len;
			break;
		}

		/* Value read from eeprom should now be in req.data[0].
		 */
		eeprom.val = req.data[0];
#ifdef I2C_DEBUG
		printk ("EEPROM read addr %03X = %02X\n",
		        eeprom.addr, eeprom.val);
#endif

		if (copy_to_user((void *)arg, &eeprom,
		                 sizeof(eeprom_request_t))) {
			rc = -EFAULT;
			break;
		}

		rc = 0;
		break;

	case I2C_CMD_EEPROM_WR:
		if (copy_from_user (&eeprom, (eeprom_request_t *)arg,
				    sizeof(eeprom_request_t))) {
			rc = -EFAULT;
			break;
		}
#ifdef I2C_DEBUG
		printk ("EERPROM write addr %03X: val %02X\n",
		        eeprom.addr, eeprom.val);
#endif

		/* Get the value of the control register.
		 * Set current address (internal pointer in the x40430)
		 * to 0x1ff.
		 */
		req.addr[0] = 0x59;
		req.alen    = 1;
		req.data[0] = 0xff;
		req.dlen    = 1;

		if ((len = si2c_write (&req)) < 0) {
			rc = len;
			break;
		}

		/* A simple read will now extract the control register value.
		 */
		req.addr[0] = 0x59;
		req.alen    = 1;
		req.dlen    = 1;

		if ((len = si2c_read (&req)) < 0) {
			rc = len;
			break;
		}

		/* Control reg should now be in req.data[0].
		 */
		ctrl_reg = req.data[0];
#ifdef I2C_DEBUG
		printk ("ctrl_reg = %02X\n", ctrl_reg);
#endif

		/* Are any of the eeprom blocks write protected?
		 */
		if (ctrl_reg & 0x18)
		{
			/* Yes - so reset these bits before proceeding with
			 * write to eeprom.
			 */ 
			ctrl_reg &= ~0x18;  /* reset block protect bits */
			ctrl_reg |= 0x02;   /* set write enable latch */
			ctrl_reg &= ~0x04;  /* clear register write enable latch */

			/* Set write enable latch.
			 */
			req.addr[0] = 0x59;
			req.alen    = 1;
			req.data[0] = 0xff;
			req.data[1] = 0x02;
			req.dlen    = 2;
       
			if ((len = si2c_write (&req)) < 0) {
				rc = len;
				break;
			}

			/* Set register write enable latch.
			 */
			req.addr[0] = 0x59;
			req.alen    = 1;
			req.data[0] = 0xff;
			req.data[1] = 0x06;
			req.dlen    = 2;

			if ((len = si2c_write (&req)) < 0) {
				rc = len;
				break;
			}

			/* Modify ctrl register.
			 */
			req.addr[0] = 0x59;
			req.alen    = 1;
			req.data[0] = 0xff;
			req.data[1] = ctrl_reg;
			req.dlen    = 2;

			if ((len = si2c_write (&req)) < 0) {
				rc = len;
				break;
			}

			/* The write (above) is an operation on NV memory.
			 * These can take some time (~5ms), and the device
			 * will not respond to further I2C messages till
			 * it's completed the write.
			 * So poll device for an I2C acknowledge.
			 * When we get one we know we can continue with other
			 * operations.
			 */
			req.addr[0] = 0x50;
			req.alen    = 1;
			req.data[0] = 0;
			req.dlen    = 1;
            
			for (i = 0; i < MAX_ACKNOWLEDGE_POLLS; i++) {
				if(si2c_read(&req) == req.dlen)
					break;                  /* got ack */
				set_current_state(TASK_INTERRUPTIBLE);
				schedule_timeout((HZ+199)/200); /* 5ms wait */
			}

			if (i == MAX_ACKNOWLEDGE_POLLS) {
				printk("EEPROM poll acknowledge failed\n");
				rc = -EIO;
				break;
			}
		}
		/* Is the write enable latch on?
		 */
		else if (!(ctrl_reg & 0x02)) {
			/* No!
			 * Set write enable latch.
			 */
			req.addr[0] = 0x59;
			req.alen    = 1;
			req.data[0] = 0xff;
			req.data[1] = 0x02;
			req.dlen    = 2;

			if ((len = si2c_write (&req)) < 0) {
				rc = len;
				break;
			}
		}

		/* Write is enabled ... now write eeprom value.
		 */
		req.addr[0] = 0x50 | ((eeprom.addr >> 8) & 1);
		req.alen    = 1;
		req.data[0] = eeprom.addr & 0xff;
		req.data[1] = eeprom.val;
		req.dlen    = 2;

		if ((len = si2c_write (&req)) < 0) {
			rc = len;
			break;
		}

		/* The write (above) is an operation on NV memory.
		 * These can take some time (~5ms), and the device
		 * will not respond to further I2C messages till
		 * it's completed the write.
		 * So poll device for an I2C acknowledge.
		 * When we get one we know we can continue with other
		 * operations.
		 */
		req.addr[0] = 0x50;
		req.alen    = 1;
		req.data[0] = 0;
		req.dlen    = 1;

		for (i = 0; i < MAX_ACKNOWLEDGE_POLLS; i++) {
			if(si2c_read(&req) == req.dlen)
				break;                      /* got ack */
			set_current_state(TASK_INTERRUPTIBLE);
			schedule_timeout((HZ+199)/200); /* 5ms wait */           
		}

		if (i == MAX_ACKNOWLEDGE_POLLS) {
			printk("EEPROM poll acknowledge failed\n");
			rc = -EIO;
			break;
		}

		rc = 0;
		break;

#endif	/* CONFIG_ICU862 */

	}

	up (&i2c_sem);

	return (rc);
}

/**********************************************************************/

static void
si2c_interrupt (void *dev_id)
{
	volatile i2c8xx_t *i2c = (i2c8xx_t *)dev_id;

#if (I2C_DEBUG > 2)
	printk ("si2c_interrupt(dev_id=%p)\n", dev_id);
#endif

	/* Chip errata, clear enable.
	 */
	i2c->i2c_i2mod = 0;

	/* Clear interrupt.
	 */
	i2c->i2c_i2cer = 0xff;

	/* Get me going again.
	 */
	wake_up_interruptible (&iic_wait);
}

/**********************************************************************/


static int
si2c_shutdown (void)
{
	volatile i2c8xx_t *i2c = si2c_data.i2c;

	/* Shut down I2C.
	*/
	i2c->i2c_i2mod = 0;
	i2c->i2c_i2cmr = 0;
	i2c->i2c_i2cer = 0xFF;

	return (0);
}

/**********************************************************************/


static void
si2c_reset_params (volatile iic_t *iip)
{
	iip->iic_tbase	= r_tbase;
	iip->iic_rbase	= r_rbase;

	iip->iic_tfcr	= SMC_EB;
	iip->iic_rfcr	= SMC_EB;

	iip->iic_mrblr	= I2C_MAX_LEN;

	iip->iic_rstate	= 0;
	iip->iic_rdp	= 0;
	iip->iic_rbptr	= 0;
	iip->iic_rbc	= 0;
	iip->iic_rxtmp	= 0;
	iip->iic_tstate	= 0;
	iip->iic_tdp	= 0;
	iip->iic_tbptr	= 0;
	iip->iic_tbc	= 0;
	iip->iic_txtmp	= 0;
}

/**********************************************************************/


static void si2c_force_close (void)
{
	if (si2c_data.reloc == 0) {
		volatile cpm8xx_t *cp = cpmp;

#ifdef I2C_DEBUG
		printk ("si2c_force_close()\n");
#endif

		cp->cp_cpcr = mk_cr_cmd (CPM_CR_CH_I2C, CPM_CR_CLOSE_RXBD) |
			      CPM_CR_FLG;

		while (cp->cp_cpcr & CPM_CR_FLG);
	}
}

/**********************************************************************/

#ifdef I2C_SCAN

/* See if an I2C address exists..
 * addr = 7 bit address, unshifted
 */
static int
si2c_tryaddress (int addr)
{
	volatile iic_t    *iip = si2c_data.iip;
	volatile i2c8xx_t *i2c = si2c_data.i2c;
	volatile cpm8xx_t *cp  = cpmp;
	volatile cbd_t *tbdf, *rbdf;
	unsigned long flags, len;
	u_char	temp[I2C_MAX_LEN];
	u_char *tb;

#if (I2C_DEBUG > 2)
	printk ("si2c_tryaddress %02X\n", addr);
#endif

	/* check for and use a microcode relocation patch */
	if (si2c_data.reloc) {
		si2c_reset_params (iip);
	}

#if (I2C_DEBUG > 2)
	if (addr == 0) {
		printk ("iip %p, dp_addr 0x%x\n"
			"iic_tbase %d, r_tbase %d\n",
			si2c_data.iip, si2c_data.dp_addr,
			iip->iic_tbase, r_tbase);
	}
#endif

	tbdf = (cbd_t *)&cp->cp_dpmem[iip->iic_tbase];
	rbdf = (cbd_t *)&cp->cp_dpmem[iip->iic_rbase];

	tb = temp;
	tb = (u_char *)(((uint)tb + 15) & ~15);

	/* do a simple read */
	tb[0] = (addr << 1) | 1;	/* device address (+ read) */
	len = 2;

	flush_dcache_range ((ulong)tb, (ulong)(tb+1));

	tbdf->cbd_bufaddr = __pa(tb);
	tbdf->cbd_datlen = len;
	tbdf->cbd_sc =
		BD_SC_READY | BD_SC_INTRPT | BD_SC_LAST |
		BD_SC_WRAP  | BD_IIC_START;

	rbdf->cbd_datlen = 0;
	rbdf->cbd_bufaddr = __pa(tb+2);
	rbdf->cbd_sc = BD_SC_EMPTY | BD_SC_WRAP;

	save_flags (flags); cli ();
	i2c->i2c_i2cmr = 0x13;	/* Enable some interupts */
	i2c->i2c_i2cer = 0xff;
	i2c->i2c_i2mod = 1;	/* Enable */
	i2c->i2c_i2com = 0x81;	/* Start master */

#if (I2C_DEBUG > 2)
	printk ("about to sleep\n");
#endif

	/* wait for I2C transfer */
	interruptible_sleep_on (&iic_wait);
	restore_flags (flags);
	if (signal_pending (current)) {
		return (-EIO);
	}

#if (I2C_DEBUG > 2)
	printk ("back from sleep\n");
#endif

	if (tbdf->cbd_sc & BD_SC_NAK) {
#if (I2C_DEBUG > 2)
		printk ("I2C try; no ack\n");
#endif
		return 0;
	}

	if (tbdf->cbd_sc & BD_SC_READY) {
		printk ("I2C try; complete but tbuf ready\n");
	}

	return 1;
}

#endif	/* I2C_SCAN */

/**********************************************************************/
