/*
 * Handle mapping of the flash memory access routines 
 * on Cambridge Boradband ICU862 Board.
 *
 * based on tqm8xxl.c
 *
 * Copyright(C) 2001
 * Wolfgang Denk, DENX Software Engineering, <wd@denx.de>
 *
 * This code is GPLed
 * 
 */

/*
 * The ICU862 uses 4 x AMD AM29LV033C (32 Mbit)
 * for a total of 16 MB flash in one bank
 *
 * Thus we have to chose:
 * - Support 32-bit buswidth => CONFIG_MTD_CFI_B4
 * - Support 4-chip flash interleave => CONFIG_MTD_CFI_I4
 */
#include <linux/config.h>
#include <linux/module.h>
#include <linux/types.h>
#include <linux/kernel.h>
#include <asm/io.h>

#include <linux/mtd/mtd.h>
#include <linux/mtd/map.h>
#include <linux/mtd/partitions.h>

#undef	DEBUG

#ifdef  DEBUG
# define debugk(fmt,args...)	printk(fmt ,##args)
#else
# define debugk(fmt,args...)
#endif


#define FLASH_BANK_MAX 2

/* trivial struct to describe partition information */
struct mtd_part_def
{
	int nums;
	unsigned char *type;
	struct mtd_partition* mtd_part;
};

static struct mtd_info* mtd_banks[FLASH_BANK_MAX];
static struct map_info* map_banks[FLASH_BANK_MAX];
static struct mtd_part_def part_banks[FLASH_BANK_MAX];
static unsigned long num_banks;
static unsigned long start_scan_addr;

static __u8 icu862_read8(struct map_info *map, unsigned long ofs)
{
	return *((__u8 *)(map->map_priv_1 + ofs));
}

static __u16 icu862_read16(struct map_info *map, unsigned long ofs)
{
	return *((__u16 *)(map->map_priv_1 + ofs));
}

static __u32 icu862_read32(struct map_info *map, unsigned long ofs)
{
	return *((__u32 *)(map->map_priv_1 + ofs));
}

static void icu862_copy_from(struct map_info *map,
		       void *to, unsigned long from, ssize_t len)
{
	memcpy_fromio(to, (void *)(map->map_priv_1 + from), len);
}

static void icu862_write8(struct map_info *map, __u8 d, unsigned long adr)
{
	*((__u8 *)(map->map_priv_1 + adr)) = d;
}

static void icu862_write16(struct map_info *map, __u16 d, unsigned long adr)
{
	*((__u16 *)( map->map_priv_1 + adr)) = d;
}

static void icu862_write32(struct map_info *map, __u32 d, unsigned long adr)
{
	*((__u32 *)(map->map_priv_1 + adr)) = d;
}

static void icu862_copy_to(struct map_info *map,
		     unsigned long to, const void *from, ssize_t len)
{
	memcpy_toio((void *)(map->map_priv_1 + to), from, len);
}

struct map_info icu862_map = {
	name:		"ICU862",
	/* size:	WINDOW_SIZE, */
	buswidth:	4,
	read8:		icu862_read8,
	read16:		icu862_read16,
	read32:		icu862_read32,
	copy_from:	icu862_copy_from,
	write8:		icu862_write8,
	write16:	icu862_write16,
	write32:	icu862_write32,
	copy_to:	icu862_copy_to
};

/*
 * The following defines the partition layout of ICU862 boards.
 *
 * See include/linux/mtd/partitions.h for definition of the
 * mtd_partition structure.
 *
 * The *_max_flash_size is the maximum possible mapped flash size
 * which is not necessarily the actual flash size. It must correspond
 * to the value specified in the mapping definition defined by the
 * "struct map_desc *_io_desc" for the corresponding machine.
 */

#ifdef CONFIG_MTD_PARTITIONS
/* Currently, ICU862 has upto 64 MB flash */
static unsigned long icu862_max_flash_size = 64 << 20;

/* partition definition for first (and only) flash bank
 * also ref. to "drivers/char/flash_config.c" 
 * (we split the first partition here into 1 MB kernel + 3 MB ramdisk)
 */
static struct mtd_partition icu862_partitions[] = {
    {
	name:	"kernel",			/* Linux kernel image	*/
	offset:	0,
	size:	1 << 20,			/*  0 ...  4 MB => 1 MB	*/
/*	mask_flags: MTD_WRITEABLE,		/ * force read-only	*/
    },
    {
	name:	"initrd",			/* Ramdisk image	*/
	offset:	1 << 20,
	size:	3 << 20,			/*  1 ...  4 MB => 3 MB	*/
    },
    {
	name:	"jffs",				/* JFFS(2) filesystem	*/
	offset:	4 << 20,
	size:	4 << 20,			/*  4 ...  8 MB => 4 MB	*/
    },
    {
	name:	"cramfs",			/* cramfs filesystem	*/
	offset:	8 << 20,
	size:	4 << 20,			/*  8 ... 12 MB => 4 MB	*/
    },
    {
	name:   "jffs2",			/* JFFS(2) filesystem	*/
	offset: 12 << 20,
	size:	 3 << 20,			/* 12 ... 15 MB => 3 MB	*/
    },
    {
	name:	"ppcboot",			/* PPCBoot Firmware	*/
	offset:	15 << 20,
	size:	256 << 10,			/* 256 kB		*/
/*	mask_flags: MTD_WRITEABLE,		/ * force read-only	*/
    },
    {
	name:	"environment",			/* PPCBoot Environment	*/
	offset:	((15 << 20) + (256 << 10)),
	size:	256 << 10,			/* 256 kB		*/
    },
    {
    	name:	"spare",			/* Unused		*/
	offset:	((15 << 20) + (512 << 10)),
	size:	512 << 10,			/* 512 kB		*/
    },
};

#if 0	/* there is only one bank */
/* partition definition for second flash bank */
static struct mtd_partition icu862_fs_partitions[] = {
    {
    }
};
#endif
#endif	/* CONFIG_MTD_PARTITIONS */

#define NB_OF(x)  (sizeof(x)/sizeof(x[0]))

int __init init_icu_mtd(void)
{
	int idx = 0, ret = 0;
	unsigned long flash_addr, flash_size, mtd_size = 0;
	/* pointer to ICU862 board info data */
	bd_t *bd = (bd_t *)__res;

	flash_addr = bd->bi_flashstart;
	flash_size = bd->bi_flashsize;

	/* request maximum flash size address space */
	start_scan_addr = (unsigned long)ioremap(flash_addr, flash_size);
	if (!start_scan_addr) {
		printk("%s: Failed to ioremap address: 0x%x\n",
			__FUNCTION__, flash_addr);
		return -EIO;
	}

	for(idx = 0 ; idx < FLASH_BANK_MAX ; idx++) {
		if (mtd_size >= flash_size)
			break;
		
		debugk ("%s: chip probing count %d\n", __FUNCTION__, idx);
		
		map_banks[idx] = (struct map_info *)kmalloc(sizeof(struct map_info),
							    GFP_KERNEL);
		if (map_banks[idx] == NULL)
		{
			ret = -ENOMEM;
			goto error_mem;
		}
		memset((void *)map_banks[idx], 0, sizeof(struct map_info));
		map_banks[idx]->name = (char *)kmalloc(16, GFP_KERNEL);
		if (map_banks[idx]->name == NULL)
		{
			ret = -ENOMEM;
			goto error_mem;
		}
		memset((void *)map_banks[idx]->name, 0, 16);
		
		sprintf(map_banks[idx]->name, "ICU862 Bank %d", idx);
		map_banks[idx]->buswidth  = 4;
		map_banks[idx]->read8     = icu862_read8;
		map_banks[idx]->read16    = icu862_read16;
		map_banks[idx]->read32    = icu862_read32;
		map_banks[idx]->copy_from = icu862_copy_from;
		map_banks[idx]->write8    = icu862_write8;
		map_banks[idx]->write16   = icu862_write16;
		map_banks[idx]->write32   = icu862_write32;
		map_banks[idx]->copy_to   = icu862_copy_to;
		map_banks[idx]->map_priv_1= 
			start_scan_addr + ((idx > 0) ? 
			(mtd_banks[idx-1] ? mtd_banks[idx-1]->size : 0) : 0);
		
		/* start to probe flash chips */
		mtd_banks[idx] = do_map_probe("cfi_probe", map_banks[idx]);
		if (mtd_banks[idx]) {
			mtd_banks[idx]->module = THIS_MODULE;
			mtd_size += mtd_banks[idx]->size;
			num_banks++;
			debugk ("%s: bank%d, name:%s, size:%dbytes \n",
				__FUNCTION__,
				num_banks, 
				mtd_banks[idx]->name,
				mtd_banks[idx]->size);
		}
	}

	/* no supported flash chips found */
	if (!num_banks) {
		printk("ICU862: No supported flash chips found!\n");
		ret = -ENXIO;
		goto error_mem;
	}

#ifdef CONFIG_MTD_PARTITIONS
	/*
	 * Select static partition definitions
	 */
	part_banks[0].mtd_part = icu862_partitions;
	part_banks[0].type = "static image";
	part_banks[0].nums = NB_OF(icu862_partitions);

#if 0	/* there is only one bank */
	part_banks[1].mtd_part = icu862_fs_partitions;
	part_banks[1].type = "static file system";
	part_banks[1].nums = NB_OF(icu862_fs_partitions);
#endif

	for(idx = 0; idx < num_banks ; idx++) {
		if (part_banks[idx].nums == 0) {
		    printk (KERN_NOTICE
			    "ICU862 flash bank %d: no partition info available, "
			    "registering whole device\n",
				idx);
		    add_mtd_device(mtd_banks[idx]);
		} else {
		    printk (KERN_NOTICE 
			    "ICU862 flash bank %d: Using %s partition definition\n",
				idx,
				part_banks[idx].type);
		    add_mtd_partitions (mtd_banks[idx],
					part_banks[idx].mtd_part, 
					part_banks[idx].nums);
		}
	}
#else
	printk (KERN_NOTICE "ICU862 flash: registering %d flash banks at once\n",
		num_banks);

	for(idx = 0 ; idx < num_banks ; idx++) {
		add_mtd_device(mtd_banks[idx]);
	}
#endif
	return 0;
error_mem:
	for(idx = 0 ; idx < FLASH_BANK_MAX ; idx++) {
		if (map_banks[idx] != NULL) {
			if (map_banks[idx]->name != NULL) {
				kfree(map_banks[idx]->name);
				map_banks[idx]->name = NULL;
			}
			kfree(map_banks[idx]);
			map_banks[idx] = NULL;
		}
	}
error:
	iounmap((void *)start_scan_addr);
	return ret;
}

static void __exit cleanup_icu_mtd(void)
{
	unsigned int idx = 0;
	for(idx = 0 ; idx < num_banks ; idx++) {
		/* destroy mtd_info previously allocated */
		if (mtd_banks[idx]) {
			del_mtd_partitions(mtd_banks[idx]);
			map_destroy(mtd_banks[idx]);
		}
		/* release map_info not used anymore */
		kfree(map_banks[idx]->name);
		kfree(map_banks[idx]);
	}
	if (start_scan_addr) {
		iounmap((void *)start_scan_addr);
		start_scan_addr = 0;
	}
}

module_init(init_icu_mtd);
module_exit(cleanup_icu_mtd);

MODULE_LICENSE("GPL");
MODULE_AUTHOR("Wolfgang Denk <wd@denx.de>");
MODULE_DESCRIPTION("MTD map driver for ICU862 boards");
