/*
 * Handle mapping of the flash memory access routines 
 * on mvSENSOR based devices.
 *
 * $Id: mvsensor.c,v 1.1.1.1 2002/06/26 15:03:14 hg Exp $
 *
 * based on rpxlite.c
 *
 * Copyright(C) 2001 Kirk Lee <kirk@hpc.ee.ntu.edu.tw>
 *
 * This code is GPLed
 * 
 */

/*
 * mvSENSOR series has following flash memory organisation:
 *	| capacity |	| chip type |	| bank0 |	| bank1 |
 *	   4MiB            4Mx16       4MiB           0 
 * Thus, we choose CONFIG_MTD_CFI_I1 & CONFIG_MTD_CFI_B2 at 
 * kernel configuration.
 */
#include <linux/config.h>
#include <linux/module.h>
#include <linux/types.h>
#include <linux/kernel.h>
#include <asm/io.h>

#include <linux/mtd/mtd.h>
#include <linux/mtd/map.h>
#include <linux/mtd/partitions.h>

#undef  DEBUG

#ifdef  DEBUG
# define debugk(fmt,args...)	printk(fmt ,##args)
#else
# define debugk(fmt,args...)
#endif


#define FLASH_BANK_MAX 1

/* trivial struct to describe partition information */
struct mtd_part_def
{
	int nums;
	unsigned char *type;
	struct mtd_partition* mtd_part;
};

static struct mtd_info* mtd_banks[FLASH_BANK_MAX];
static struct map_info* map_banks[FLASH_BANK_MAX];
static struct mtd_part_def part_banks[FLASH_BANK_MAX];
static unsigned long num_banks;
static unsigned long start_scan_addr;

__u8 mvs_read8(struct map_info *map, unsigned long ofs)
{
	return *((__u8 *)(map->map_priv_1 + ofs));
}

__u16 mvs_read16(struct map_info *map, unsigned long ofs)
{
	return *((__u16 *)(map->map_priv_1 + ofs));
}

__u32 mvs_read32(struct map_info *map, unsigned long ofs)
{
	return *((__u32 *)(map->map_priv_1 + ofs));
}

void mvs_copy_from(struct map_info *map,
		       void *to, unsigned long from, ssize_t len)
{
	memcpy_fromio(to, (void *)(map->map_priv_1 + from), len);
}

void mvs_write8(struct map_info *map, __u8 d, unsigned long adr)
{
	*((__u8 *)(map->map_priv_1 + adr)) = d;
}

void mvs_write16(struct map_info *map, __u16 d, unsigned long adr)
{
	*((__u16 *)( map->map_priv_1 + adr)) = d;
}

void mvs_write32(struct map_info *map, __u32 d, unsigned long adr)
{
	*((__u32 *)(map->map_priv_1 + adr)) = d;
}

void mvs_copy_to(struct map_info *map,
		     unsigned long to, const void *from, ssize_t len)
{
	memcpy_toio((void *)(map->map_priv_1 + to), from, len);
}

struct map_info mvs_map = {
	name:		"mvSENSOR",
	/* size:	WINDOW_SIZE, */
	buswidth:	2,
	read8:		mvs_read8,
	read16:		mvs_read16,
	read32:		mvs_read32,
	copy_from:	mvs_copy_from,
	write8:		mvs_write8,
	write16:	mvs_write16,
	write32:	mvs_write32,
	copy_to:	mvs_copy_to
};

/*
 * Here are partition information for all known mvSENSOR series devices.
 * See include/linux/mtd/partitions.h for definition of the mtd_partition
 * structure.
 * 
 * The *_max_flash_size is the maximum possible mapped flash size which
 * is not necessarily the actual flash size.  It must correspond to the 
 * value specified in the mapping definition defined by the
 * "struct map_desc *_io_desc" for the corresponding machine.
 */

#ifdef CONFIG_MTD_PARTITIONS
static unsigned long mvs_max_flash_size = 0x00400000;
/* partition definition for first flash bank
 * also ref. to "drivers\char\flash_config.c" 
 */
static struct mtd_partition mvs_partitions[] = {
	{	/* new MVsensor */
	  name: 	"ppcboot",
	  offset: 	0x00000000,
	  size: 	0x00020000,			/* 128 kB read/write	*/
//	  mask_flags: MTD_WRITEABLE,	/* force read-only */
	},
	{
	  name: 	"kernel",
	  offset: 	0x00020000,
	  size: 	0x00080000,			/* 512 kB read/write	*/
	},
	{
	  name: 	"root fs",
	  offset: 	0x000a0000,
	  size: 	0x001e0000,			/* 1920 kB read/write	*/		
	},
	{
	  name: 	"user",
	  offset: 	0x00280000,
	  size: 	0x00180000,			/* 1536 kB read / write */ 			
	},
	{
	  name: 	"ppcboot_env",
	  offset: 	0x00008000,
	  size: 	0x00002000,
	},
};
#endif

#define NB_OF(x)  (sizeof(x)/sizeof(x[0]))

int __init init_mvs_mtd(void)
{
	int idx = 0, ret = 0;
	unsigned long flash_addr, flash_size, mtd_size = 0;
	/* pointer to mvSENSOR board info data */
	bd_t *bd = (bd_t *)__res;

	flash_addr = bd->bi_flashstart;
	flash_size = bd->bi_flashsize;

	/* request maximum flash size address space */
	start_scan_addr = (unsigned long)ioremap(flash_addr, flash_size);
	if (!start_scan_addr) {
		printk("%s: Failed to ioremap address: 0x%x\n",
			__FUNCTION__, flash_addr);
		return -EIO;
	}

	for(idx = 0 ; idx < FLASH_BANK_MAX ; idx++) {
		if (mtd_size >= flash_size)
			break;
		
		debugk ("%s: chip probing count %d\n", __FUNCTION__, idx);
		
		map_banks[idx] = (struct map_info *)kmalloc(sizeof(struct map_info),
							    GFP_KERNEL);
		if (map_banks[idx] == NULL)
		{
			ret = -ENOMEM;
			goto error_mem;
		}
		memset((void *)map_banks[idx], 0, sizeof(struct map_info));
		map_banks[idx]->name = (char *)kmalloc(16, GFP_KERNEL);
		if (map_banks[idx]->name == NULL)
		{
			ret = -ENOMEM;
			goto error_mem;
		}
		memset((void *)map_banks[idx]->name, 0, 16);
		
		sprintf(map_banks[idx]->name, "mvSENSOR Bank %d", idx);
		map_banks[idx]->buswidth  = 2;
		map_banks[idx]->read8     = mvs_read8;
		map_banks[idx]->read16    = mvs_read16;
		map_banks[idx]->read32    = mvs_read32;
		map_banks[idx]->copy_from = mvs_copy_from;
		map_banks[idx]->write8    = mvs_write8;
		map_banks[idx]->write16   = mvs_write16;
		map_banks[idx]->write32   = mvs_write32;
		map_banks[idx]->copy_to   = mvs_copy_to;
		map_banks[idx]->map_priv_1= 
			start_scan_addr + ((idx > 0) ? 
			(mtd_banks[idx-1] ? mtd_banks[idx-1]->size : 0) : 0);
		
		/* start to probe flash chips */
		mtd_banks[idx] = do_map_probe("cfi_probe", map_banks[idx]);
		if (mtd_banks[idx]) {
			mtd_banks[idx]->module = THIS_MODULE;
			mtd_size += mtd_banks[idx]->size;
			num_banks++;
			debugk ("%s: bank%d, name:%s, size:%dbytes \n",
				__FUNCTION__,
				num_banks, 
				mtd_banks[idx]->name,
				mtd_banks[idx]->size);
		}
	}

	/* no supported flash chips found */
	if (!num_banks) {
		printk("mvSENSOR: No support flash chips found!\n");
		ret = -ENXIO;
		goto error_mem;
	}

#ifdef CONFIG_MTD_PARTITIONS
	/*
	 * Select static partition definitions
	 */
	part_banks[0].mtd_part = mvs_partitions;
	part_banks[0].type = "static image";
	part_banks[0].nums = NB_OF(mvs_partitions);

	for(idx = 0; idx < num_banks ; idx++) {
		if (part_banks[idx].nums == 0) {
		    printk (KERN_NOTICE
			    "mvSENSOR flash bank %d: no partition info available, "
			    "registering whole device\n",
				idx);
		    add_mtd_device(mtd_banks[idx]);
		} else {
		    printk (KERN_NOTICE 
			    "mvSENSOR flash bank %d: Using %s partition definition\n",
				idx,
				part_banks[idx].type);
		    add_mtd_partitions (mtd_banks[idx],
					part_banks[idx].mtd_part, 
					part_banks[idx].nums);
		}
	}
#else
	printk (KERN_NOTICE "mvSENSOR flash: registering %d flash banks at once\n",
		num_banks);

	for(idx = 0 ; idx < num_banks ; idx++) {
		add_mtd_device(mtd_banks[idx]);
	}
#endif
	return 0;
error_mem:
	for(idx = 0 ; idx < FLASH_BANK_MAX ; idx++) {
		if (map_banks[idx] != NULL) {
			if (map_banks[idx]->name != NULL) {
				kfree(map_banks[idx]->name);
				map_banks[idx]->name = NULL;
			}
			kfree(map_banks[idx]);
			map_banks[idx] = NULL;
		}
	}
error:
	iounmap((void *)start_scan_addr);
	return ret;
}

static void __exit cleanup_mvs_mtd(void)
{
	unsigned int idx = 0;
	for(idx = 0 ; idx < num_banks ; idx++) {
		/* destroy mtd_info previously allocated */
		if (mtd_banks[idx]) {
			del_mtd_partitions(mtd_banks[idx]);
			map_destroy(mtd_banks[idx]);
		}
		/* release map_info not used anymore */
		kfree(map_banks[idx]->name);
		kfree(map_banks[idx]);
	}
	if (start_scan_addr) {
		iounmap((void *)start_scan_addr);
		start_scan_addr = 0;
	}
}

module_init(init_mvs_mtd);
module_exit(cleanup_mvs_mtd);

MODULE_LICENSE("GPL");
MODULE_AUTHOR("Kirk Lee <kirk@hpc.ee.ntu.edu.tw> and MATRIX Vision GmbH");
MODULE_DESCRIPTION("MTD map driver for mvSENSOR boards");
